from typing import Annotated

from pydantic import Field, model_validator

from mex.common.models import BaseModel


class Value(BaseModel):
    """Model class for Values (for claims)."""

    text: str | None = None
    language: str | None = None


class DataValue(BaseModel):
    """Model class for Data Values (for claims)."""

    value: Value

    @model_validator(mode="before")
    @classmethod
    def transform_strings_to_dict(
        cls, values: dict[str, str | dict[str, str]]
    ) -> dict[str, dict[str, str | None]] | dict[str, str | dict[str, str]]:
        """Transform string and null value to a dict for parsing.

        Args:
            values: values that needs to be parsed

        Returns:
            resulting dict
        """
        value = values.get("value")
        if value is None or isinstance(value, str):
            return {"value": {"language": None, "text": value}}
        return values


class Mainsnak(BaseModel):
    """Model class for Mainsnack (for claims)."""

    datavalue: DataValue


class Claim(BaseModel):
    """Model class a Claim."""

    mainsnak: Mainsnak


class Claims(BaseModel):
    """model class for Claims."""

    website: Annotated[list[Claim], Field(alias="P856")] = []
    isni_id: Annotated[list[Claim], Field(alias="P213")] = []
    ror_id: Annotated[list[Claim], Field(alias="P6782")] = []
    official_name: Annotated[list[Claim], Field(alias="P1448")] = []
    short_name: Annotated[list[Claim], Field(alias="P1813")] = []
    native_label: Annotated[list[Claim], Field(alias="P1705")] = []
    gepris_id: Annotated[list[Claim], Field(alias="P4871")] = []
    gnd_id: Annotated[list[Claim], Field(alias="P227")] = []
    viaf_id: Annotated[list[Claim], Field(alias="P214")] = []


class Label(BaseModel):
    """Model class for single Label."""

    language: str | None = None
    value: str


class Labels(BaseModel):
    """Model class for Labels."""

    de: Label | None = None
    en: Label | None = None
    multiple: Annotated[Label | None, Field(alias="mul")] = None


class Alias(BaseModel):
    """Model class for single alias."""

    language: str
    value: str


class Aliases(BaseModel):
    """Model class for aliases."""

    de: list[Alias] = []
    en: list[Alias] = []


class WikidataOrganization(BaseModel):
    """Model class for Wikidata sources."""

    identifier: Annotated[str, Field(alias="id")]
    labels: Labels
    claims: Claims
    aliases: Aliases



import pytest

from mex.common.context import SingleSingletonStore, SingletonStore


class Parent:
    pass


class Child(Parent):
    pass


def test_singleton_store() -> None:
    store = SingletonStore["Parent"]()
    parent = store.load(Parent)

    popped = store.pop(Parent)
    assert popped is parent

    new_parent = store.load(Parent)
    assert new_parent is not parent

    store.reset()

    assert not store._instances_by_class


def test_single_singleton_store() -> None:
    store = SingleSingletonStore["Parent"]()
    parent = store.load(Parent)

    with pytest.raises(RuntimeError, match="is not a parent class of loaded class"):
        store.load(Child)

    assert parent is store.load(Parent)

    store.push(Child())

    child = store.load(Child)

    assert child is store.load(Parent)

    store.reset()

    assert store._singleton is None



from typing import Any

import pytest

from mex.common.orcid.models import (
    OrcidEmail,
    OrcidEmails,
    OrcidFamilyName,
    OrcidGivenNames,
    OrcidIdentifier,
    OrcidName,
    OrcidPerson,
    OrcidRecord,
)
from mex.common.orcid.transform import transform_orcid_person_to_mex_person
from mex.common.types import MergedPrimarySourceIdentifier


@pytest.mark.parametrize(
    ("orcid_person", "expected_mex_person"),
    [
        (
            OrcidRecord(
                orcid_identifier=OrcidIdentifier(
                    path="0009-0004-3041-5706",
                    uri="https://orcid.org/0009-0004-3041-5706",
                ),
                person=OrcidPerson(
                    orcid_identifier="0009-0004-3041-5706",
                    emails=OrcidEmails(email=[OrcidEmail(email=["test@example.com"])]),
                    name=OrcidName(
                        given_names=OrcidGivenNames(value="Josiah"),
                        family_name=OrcidFamilyName(value="Carberry"),
                        visibility="public",
                    ),
                ),
            ),
            {
                "hadPrimarySource": "Naj2hOJq9FNRkkMWa5Qd0",
                "identifierInPrimarySource": "0009-0004-3041-5706",
                "affiliation": [],
                "email": ["test@example.com"],
                "familyName": ["Carberry"],
                "fullName": ["Carberry, Josiah"],
                "givenName": ["Josiah"],
                "isniId": [],
                "memberOf": [],
                "orcidId": ["https://orcid.org/0009-0004-3041-5706"],
                "entityType": "ExtractedPerson",
                "identifier": "eLFbvlVkwRgGxLnS8RywZ3",
                "stableTargetId": "ccM55btPUrNtYKSLX8cNQP",
            },
        ),
        (
            OrcidRecord(
                orcid_identifier=OrcidIdentifier(
                    path="0000-0002-9876-5432",
                    uri="https://orcid.org/0000-0002-9876-5432",
                ),
                person=OrcidPerson(
                    orcid_identifier="0000-0002-9876-5432",
                    emails=OrcidEmails(email=[OrcidEmail(email=[])]),
                    name=OrcidName(
                        given_names=OrcidGivenNames(value="John"),
                        family_name=OrcidFamilyName(value="Doe"),
                        visibility="public",
                    ),
                ),
            ),
            {
                "hadPrimarySource": "Naj2hOJq9FNRkkMWa5Qd0",
                "identifierInPrimarySource": "0000-0002-9876-5432",
                "affiliation": [],
                "email": [],
                "familyName": ["Doe"],
                "fullName": ["Doe, John"],
                "givenName": ["John"],
                "isniId": [],
                "memberOf": [],
                "orcidId": ["https://orcid.org/0000-0002-9876-5432"],
                "entityType": "ExtractedPerson",
                "identifier": "cwOd1omDQ8ePMflUVfXmH6",
                "stableTargetId": "cPgmFF42nbFoEMWtuHaQw7",
            },
        ),
    ],
    ids=[
        "existing person with email",
        "existing person with no email",
    ],
)
def test_transform_orcid_person_to_mex_person(
    orcid_person: OrcidRecord,
    expected_mex_person: dict[str, Any],
    extracted_primary_source_ids: dict[str, MergedPrimarySourceIdentifier],
) -> None:
    mex_person = transform_orcid_person_to_mex_person(
        orcid_person, extracted_primary_source_ids["orcid"]
    )
    assert mex_person.model_dump() == expected_mex_person



from mex.common.models import (
    ADDITIVE_MODEL_CLASSES,
    BASE_MODEL_CLASSES_BY_NAME,
    EXTRACTED_MODEL_CLASSES,
    PREVENTIVE_MODEL_CLASSES,
    RULE_SET_REQUEST_CLASSES,
    RULE_SET_RESPONSE_CLASSES,
    SUBTRACTIVE_MODEL_CLASSES,
)
from mex.common.types import Identifier, MergedPrimarySourceIdentifier


def test_all_rules_are_defined() -> None:
    stem_types = sorted(c.stemType for c in EXTRACTED_MODEL_CLASSES)
    for lookup in (
        ADDITIVE_MODEL_CLASSES,
        PREVENTIVE_MODEL_CLASSES,
        RULE_SET_REQUEST_CLASSES,
        RULE_SET_RESPONSE_CLASSES,
        SUBTRACTIVE_MODEL_CLASSES,
    ):
        assert sorted(c.stemType for c in lookup) == stem_types


def test_additive_models_define_same_fields_as_base_model() -> None:
    for additive_rule in ADDITIVE_MODEL_CLASSES:
        base_model_name = "Base" + additive_rule.stemType
        base_model = BASE_MODEL_CLASSES_BY_NAME[base_model_name]
        expected_fields = {"entityType", *base_model.model_fields}
        assert set(additive_rule.model_fields) == expected_fields


def test_additive_models_have_no_required_fields() -> None:
    for additive_model in ADDITIVE_MODEL_CLASSES:
        model = additive_model()
        assert model.model_dump(exclude_unset=True) == {}


def test_subtractive_models_define_same_fields_as_base_model() -> None:
    for subtractive_rule in SUBTRACTIVE_MODEL_CLASSES:
        base_model_name = "Base" + subtractive_rule.stemType
        base_model = BASE_MODEL_CLASSES_BY_NAME[base_model_name]
        expected_fields = {"entityType", *base_model.model_fields}
        assert set(subtractive_rule.model_fields) == expected_fields


def test_subtractive_models_have_no_required_fields() -> None:
    for subtractive_model in SUBTRACTIVE_MODEL_CLASSES:
        model = subtractive_model()
        assert model.model_dump(exclude_unset=True) == {}


def test_preventive_models_define_same_fields_as_base_model() -> None:
    for preventive_rule in PREVENTIVE_MODEL_CLASSES:
        base_model_name = "Base" + preventive_rule.stemType
        base_model = BASE_MODEL_CLASSES_BY_NAME[base_model_name]
        expected_fields = {"entityType", *base_model.model_fields}
        assert set(preventive_rule.model_fields) == expected_fields


def test_preventive_models_define_all_fields_as_correct_type() -> None:
    for preventive_rule in PREVENTIVE_MODEL_CLASSES:
        for field_name, field_info in preventive_rule.model_fields.items():
            if field_name == "entityType":
                assert field_info.default == preventive_rule.__name__
            else:
                assert field_info.annotation == list[MergedPrimarySourceIdentifier]
                assert field_info.is_required() is False
                assert field_info.default == []


def test_rule_set_request_models_have_no_required_fields() -> None:
    for rule_set_request in RULE_SET_REQUEST_CLASSES:
        model = rule_set_request()
        assert model.model_dump(exclude_unset=True) == {}


def test_rule_set_response_models_only_require_stable_target_id() -> None:
    for rule_set_response in RULE_SET_RESPONSE_CLASSES:
        model = rule_set_response(stableTargetId=Identifier.generate(seed=1))
        assert model.model_dump(exclude_unset=True) == {
            "stableTargetId": "bFQoRhcVH5DHUr"
        }



from collections.abc import Iterable, Mapping
from itertools import chain
from typing import Literal, cast, overload

from pydantic_core import ValidationError

from mex.common.exceptions import MergingError
from mex.common.fields import MERGEABLE_FIELDS_BY_CLASS_NAME
from mex.common.logging import logger
from mex.common.merged.types import (
    SourceAndValueIter,
    SourceAndValueList,
    SourceList,
    ValueList,
)
from mex.common.models import (
    MERGED_MODEL_CLASSES_BY_NAME,
    MEX_PRIMARY_SOURCE_STABLE_TARGET_ID,
    PREVIEW_MODEL_CLASSES_BY_NAME,
    RULE_SET_REQUEST_CLASSES_BY_NAME,
    AnyExtractedModel,
    AnyMergedModel,
    AnyPreviewModel,
    AnyRuleSetRequest,
    AnyRuleSetResponse,
)
from mex.common.transform import ensure_prefix
from mex.common.types import AnyPrimitiveType, AnyValidation, Identifier, Validation
from mex.common.utils import ensure_list


def _collect_extracted_values(
    field: str,
    extracted_items: Iterable[AnyExtractedModel],
) -> SourceAndValueList:
    """Collect values from extracted items for a specific field.

    Args:
        field: Name of the field to extract values from
        extracted_items: Iterable of extracted model instances

    Returns:
        List of (source, value) tuples from the extracted items
    """
    return [
        (extracted_item.hadPrimarySource, value)
        for extracted_item in extracted_items
        for value in cast("ValueList", ensure_list(getattr(extracted_item, field)))
    ]


def _collect_additive_values(
    field: str,
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse,
) -> SourceAndValueList:
    """Collect values from additive rules for a specific field.

    Args:
        field: Name of the field to extract values from
        rule_set: Rule set containing additive rules

    Returns:
        List of (source, value) tuples from the additive rules
    """
    return [
        (MEX_PRIMARY_SOURCE_STABLE_TARGET_ID, value)
        for value in cast("ValueList", ensure_list(getattr(rule_set.additive, field)))
    ]


def _collect_subtractive_values(
    field: str,
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse,
) -> ValueList:
    """Collect values from subtractive rules for a specific field.

    Args:
        field: Name of the field to extract values from
        rule_set: Rule set containing subtractive rules

    Returns:
        List of values that should be subtracted/removed
    """
    return cast("ValueList", ensure_list(getattr(rule_set.subtractive, field)))


def _collect_preventive_sources(
    field: str,
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse,
) -> SourceList:
    """Collect sources from preventive rules for a specific field.

    Args:
        field: Name of the field to extract sources from
        rule_set: Rule set containing preventive rules

    Returns:
        List of sources that should be prevented/blocked
    """
    return cast("SourceList", ensure_list(getattr(rule_set.preventive, field)))


def _filter_usable_values(
    possible_sources_and_values: SourceAndValueIter,
    prevented_sources: SourceList,
    subtracted_values: ValueList,
) -> ValueList:
    """Filter values by removing prevented sources and subtracted values.

    Args:
        possible_sources_and_values: Iterable of (source, value) tuples to filter
        prevented_sources: List of sources that should be excluded
        subtracted_values: List of values that should be excluded

    Returns:
        List of unique, filtered values
    """
    seen_values = set()
    usable_values = []

    for source, value in possible_sources_and_values:
        if (
            source not in prevented_sources
            and value not in subtracted_values
            and value not in seen_values
        ):
            seen_values.add(value)
            usable_values.append(value)

    return usable_values


def _apply_lenient_fallback(
    extracted_sources_and_values: SourceAndValueIter,
    additive_rule_sources_and_values: SourceAndValueIter,
    subtracted_values: ValueList,
) -> ValueList:
    """Apply lenient fallback by returning first available value from any source.

    Args:
        extracted_sources_and_values: (source, value) tuples from extracted items
        additive_rule_sources_and_values: (source, value) tuples from additive rules
        subtracted_values: Values that were subtracted/removed

    Returns:
        List containing the first available value, or empty list if none found
    """
    subtractive_rule_sources_and_values: SourceAndValueIter = (
        (MEX_PRIMARY_SOURCE_STABLE_TARGET_ID, value) for value in subtracted_values
    )
    for _, value in chain(
        extracted_sources_and_values,
        additive_rule_sources_and_values,
        subtractive_rule_sources_and_values,
    ):
        return [value]
    return []


def _pick_usable_values(
    field: str,
    extracted_items: Iterable[AnyExtractedModel],
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse,
    validation: AnyValidation,
) -> ValueList:
    """Pick usable values for a field from given extracted items and rules.

    Args:
        field: Name of the field to process
        extracted_items: Iterable of extracted model instances
        rule_set: Rule set containing additive, subtractive, and preventive rules
        validation: Validation mode (STRICT, LENIENT, or IGNORE)

    Returns:
        List of usable values after applying all rules and validation logic
    """
    # Collect values from extracted items and rules
    extracted_sources_and_values = _collect_extracted_values(field, extracted_items)
    additive_rule_sources_and_values = _collect_additive_values(field, rule_set)
    subtracted_values = _collect_subtractive_values(field, rule_set)
    prevented_sources = _collect_preventive_sources(field, rule_set)

    # Combine possible sources and filter them
    possible_sources_and_values = chain(
        extracted_sources_and_values, additive_rule_sources_and_values
    )
    usable_values = _filter_usable_values(
        possible_sources_and_values, prevented_sources, subtracted_values
    )

    # Apply lenient fallback if needed
    allow_lenient_fallback = validation is Validation.LENIENT
    if not usable_values and allow_lenient_fallback:
        return _apply_lenient_fallback(
            extracted_sources_and_values,
            additive_rule_sources_and_values,
            subtracted_values,
        )

    return usable_values


def _create_merged_dict(
    mergeable_fields: Iterable[str],
    extracted_items: Iterable[AnyExtractedModel],
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse,
    validation: AnyValidation,
) -> dict[str, list[AnyPrimitiveType] | AnyPrimitiveType]:
    """Create a merged dictionary by processing all mergeable fields.

    Args:
        mergeable_fields: Names of fields that should be processed for merging
        extracted_items: Iterable of extracted model instances
        rule_set: Rule set containing additive, subtractive, and preventive rules
        validation: Validation mode (STRICT, LENIENT, or IGNORE)

    Returns:
        Dictionary mapping field names to lists of merged values
    """
    return {
        field: _pick_usable_values(
            field,
            extracted_items,
            rule_set,
            validation,
        )
        for field in mergeable_fields
    }


def _get_merged_class(
    extracted_items: list[AnyExtractedModel],
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse | None,
    validation: AnyValidation,
) -> type[AnyMergedModel | AnyPreviewModel] | None:
    """Determine the appropriate merged model class based on validation mode.

    Args:
        extracted_items: List of extracted model instances
        rule_set: Rule set containing entity type information, or None
        validation: Validation mode determining whether to use Preview or Merged models

    Returns:
        The appropriate model class (Preview* for LENIENT, Merged* for STRICT/IGNORE),
        or None if neither extracted_items nor rule_set was given.
    """
    model_class_lookup: Mapping[str, type[AnyPreviewModel | AnyMergedModel]]
    if validation == Validation.LENIENT:
        model_prefix = "Preview"
        model_class_lookup = PREVIEW_MODEL_CLASSES_BY_NAME
    else:
        model_prefix = "Merged"
        model_class_lookup = MERGED_MODEL_CLASSES_BY_NAME

    if rule_set:
        stem_type = rule_set.stemType
    elif extracted_items:
        stem_type = extracted_items[0].stemType
    else:
        return None

    entity_type = ensure_prefix(stem_type, model_prefix)
    return model_class_lookup[entity_type]


def _ensure_rule_set(
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse | None,
    stem_type: str,
) -> AnyRuleSetRequest | AnyRuleSetResponse:
    """Ensure a rule set instance exists, creating a default one if needed.

    Args:
        rule_set: Existing rule set instance, or None to create a default
        stem_type: Entity stem type (e.g., "Person", "ContactPoint") for creation

    Returns:
        A rule set instance (either the provided one or a newly created default)

    Raises:
        KeyError: If stem_type does not correspond to a valid rule set class
    """
    if rule_set is None:
        rule_set_class_name = f"{stem_type}RuleSetRequest"
        rule_set = RULE_SET_REQUEST_CLASSES_BY_NAME[rule_set_class_name]()
    return rule_set


@overload
def create_merged_item(
    identifier: Identifier,
    extracted_items: Iterable[AnyExtractedModel],
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse | None,
    validation: Literal[Validation.LENIENT],
) -> AnyPreviewModel: ...


@overload
def create_merged_item(
    identifier: Identifier,
    extracted_items: Iterable[AnyExtractedModel],
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse | None,
    validation: Literal[Validation.STRICT],
) -> AnyMergedModel: ...


@overload
def create_merged_item(
    identifier: Identifier,
    extracted_items: Iterable[AnyExtractedModel],
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse | None,
    validation: Literal[Validation.IGNORE],
) -> AnyMergedModel | None: ...


def create_merged_item(
    identifier: Identifier,
    extracted_items: Iterable[AnyExtractedModel],
    rule_set: AnyRuleSetRequest | AnyRuleSetResponse | None,
    validation: AnyValidation,
) -> AnyPreviewModel | AnyMergedModel | None:
    """Merge a list of extracted items with a set of rules.

    Args:
        identifier: Identifier the finished merged item should have
        extracted_items: List of extracted items, can be empty
        rule_set: Rule set, with potentially empty rules
        validation: Controls how strictly the merged item needs to validate:
            - STRICT: Validates all required fields and list lengths. Returns a
                fully validated merged item or raises MergingError on validation failure
            - LENIENT: Skips validation checks and returns a "preview" merged item
                that may be missing required fields and even may be using blocked values
            - IGNORE: In case of validation errors, this mode will safely return None

    Raises:
        MergingError: When the given items cannot be merged (in STRICT mode)
        MergingError: When neither extracted nor rule items are given (in STRICT mode)

    Returns:
        Instance of a merged or preview item
    """
    # Convert extracted items from any iterable into sorted list
    extracted_items = sorted(extracted_items, key=lambda e: e.identifier)

    # Get merged class based on extracted and rule items and validation mode
    merged_class = _get_merged_class(extracted_items, rule_set, validation)

    # Bail out when neither extracted nor rule items are given
    if merged_class is None:
        if validation == Validation.STRICT:
            msg = "One of rule_set or extracted_items is required."
            raise MergingError(msg)
        logger.debug("One of rule_set or extracted_items is required.")
        return None

    # Get mergeable fields and ensure rule set instance
    fields = MERGEABLE_FIELDS_BY_CLASS_NAME[merged_class.__name__]
    rule_set = _ensure_rule_set(rule_set, merged_class.stemType)

    # Create a dictionary of merged values and set the merged identifier
    merged_dict = _create_merged_dict(fields, extracted_items, rule_set, validation)
    merged_dict["identifier"] = identifier

    # Try to parse the dictionary of merged values into a pydantic model
    try:
        return merged_class.model_validate(merged_dict)
    except ValidationError as error:
        msg = "Could not validate merged model."
        if validation == Validation.STRICT:
            raise MergingError(msg) from error
        logger.debug("%s %s:%s", msg, merged_class.__name__, identifier)
    return None



from enum import Enum
from uuid import UUID

from pydantic import UUID4

from mex.common.models import BaseModel
from mex.common.settings import BaseSettings
from mex.common.sinks.ndjson import NdjsonSink
from mex.common.types import Identifier, TemporalEntity


class DummyEnum(Enum):
    NAME = "value"


class Thing(BaseModel):
    identifier: Identifier
    str_attr: str
    enum_attr: DummyEnum | None = None
    uuid_attr: UUID4 | None = None
    ts_attr: TemporalEntity | None = None


def test_sink_load() -> None:
    settings = BaseSettings.get()

    test_models = [
        Thing(identifier=Identifier.generate(seed=1), str_attr="foo"),
        Thing(
            identifier=Identifier.generate(seed=2),
            str_attr="bar",
            enum_attr=DummyEnum.NAME,
        ),
        Thing(
            identifier=Identifier.generate(seed=3),
            str_attr="baz",
            uuid_attr=UUID(int=42, version=4),
        ),
        Thing(
            identifier=Identifier.generate(seed=4),
            str_attr="dat",
            ts_attr=TemporalEntity(2000, 1, 1),
        ),
    ]

    sink = NdjsonSink.get()
    items = list(sink.load(test_models))
    assert len(items) == len(test_models)

    with (settings.work_dir / "Thing.ndjson").open() as fh:
        output = fh.read()

    expected = """\
{{"enum_attr": null, "identifier": "{}", "str_attr": "foo", "ts_attr": null, "uuid_attr": null}}
{{"enum_attr": "value", "identifier": "{}", "str_attr": "bar", "ts_attr": null, "uuid_attr": null}}
{{"enum_attr": null, "identifier": "{}", "str_attr": "baz", "ts_attr": null, "uuid_attr": "00000000-0000-4000-8000-00000000002a"}}
{{"enum_attr": null, "identifier": "{}", "str_attr": "dat", "ts_attr": "2000-01-01", "uuid_attr": null}}
""".format(*[m.identifier for m in test_models])

    assert output == expected



from enum import Enum
from typing import Annotated, Literal

import pytest
from pydantic import Field, ValidationError, computed_field

from mex.common.identity import get_provider
from mex.common.models import BaseModel, ExtractedData
from mex.common.types import (
    ExtractedIdentifier,
    MergedIdentifier,
    MergedPrimarySourceIdentifier,
)


class Animal(Enum):
    """Dummy enum to use in tests."""

    CAT = "cat"
    DOG = "dog"


class ExtractedThingIdentifier(ExtractedIdentifier):
    """Identifier for extracted things."""


class MergedThingIdentifier(MergedIdentifier):
    """Identifier for merged thing."""


class BaseThing(BaseModel):
    """Dummy model defining some arbitrary field."""

    someField: str = "someDefault"


class ExtractedThing(BaseThing, ExtractedData):
    """Extracted version of a dummy thing model."""

    entityType: Annotated[
        Literal["ExtractedThing"], Field(alias="$type", frozen=True)
    ] = "ExtractedThing"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(self) -> ExtractedThingIdentifier:
        """Return the computed identifier for this extracted item."""
        return self._get_identifier(ExtractedThingIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(self) -> MergedThingIdentifier:  # noqa: N802
        """Return the computed stableTargetId for this extracted item."""
        return self._get_stable_target_id(MergedThingIdentifier)


def test_extracted_data_requires_dict_for_construction() -> None:
    with pytest.raises(ValidationError, match="Input should be a valid dictionary"):
        ExtractedThing.model_validate(["this is a list"])


def test_extracted_data_requires_identifier_in_primary_source() -> None:
    with pytest.raises(ValidationError, match="identifierInPrimarySource"):
        ExtractedThing(  # type: ignore[call-arg]
            hadPrimarySource=MergedPrimarySourceIdentifier.generate(seed=1),
        )


def test_extracted_data_requires_had_primary_source() -> None:
    with pytest.raises(ValidationError, match="hadPrimarySource"):
        ExtractedThing(  # type: ignore[call-arg]
            identifierInPrimarySource="0",
        )


def test_extracted_data_stores_identity_in_provider() -> None:
    thing = ExtractedThing(
        identifierInPrimarySource="12345",
        hadPrimarySource=MergedPrimarySourceIdentifier.generate(seed=12345),
    )
    assert thing.identifier
    assert thing.stableTargetId

    provider = get_provider()
    identities = provider.fetch(
        had_primary_source=thing.hadPrimarySource,
        identifier_in_primary_source=thing.identifierInPrimarySource,
    )
    assert len(identities) == 1
    assert str(thing.identifier) == identities[0].identifier
    assert str(thing.stableTargetId) == identities[0].stableTargetId



from mex.common.primary_source.extract import extract_seed_primary_sources
from mex.common.types import TextLanguage


def test_extract_seed_primary_sources() -> None:
    seed_primary_source = list(extract_seed_primary_sources())

    assert len(seed_primary_source) == 5
    assert seed_primary_source[0].model_dump() == {
        "identifier": "mex",
        "title": [{"value": "Metadata Exchange", "language": TextLanguage.EN}],
    }



import functools
import warnings
from collections.abc import Callable, Container, Generator, Iterable, Iterator, Mapping
from dataclasses import dataclass
from functools import lru_cache
from itertools import zip_longest
from random import random
from time import sleep
from types import NoneType, UnionType
from typing import (
    Annotated,
    Any,
    Literal,
    ParamSpec,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from pydantic import BaseModel, TypeAdapter, ValidationError

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")


@dataclass
class GenericFieldInfo:
    """Abstraction class for unifying `FieldInfo` and `ComputedFieldInfo` objects."""

    alias: str | None
    annotation: type[Any] | None
    frozen: bool


def contains_any(base: Container[T], tokens: Iterable[T]) -> bool:
    """Check if a given base contains any of the given tokens."""
    return any(token in base for token in tokens)


def any_contains_any(bases: Iterable[Container[T] | None], tokens: Iterable[T]) -> bool:
    """Check if any of the given bases contains any of the given tokens."""
    for base in bases:
        if base is None:
            continue
        for token in tokens:
            if token in base:
                return True
    return False


def contains_only_types(field: GenericFieldInfo, *types: type) -> bool:
    """Return whether a `field` is annotated as one of the given `types`.

    Unions, lists and type annotations are checked for their inner types and only the
    non-`NoneType` types are considered for the type-check.

    Args:
        field: A `GenericFieldInfo` instance
        types: Types to look for in the field's annotation

    Returns:
        Whether the field contains any of the given types
    """
    if inner_types := list(get_inner_types(field.annotation, include_none=False)):
        return all(inner_type in types for inner_type in inner_types)
    return False


def contains_any_types(field: GenericFieldInfo, *types: type) -> bool:
    """Return whether a `field` is annotated as any of the given `types`.

    Unions, lists and type annotations are checked for their inner types and only the
    non-`NoneType` types are considered for the type-check.

    Args:
        field: A `GenericFieldInfo` instance
        types: Types to look for in the field's annotation

    Returns:
        Whether the field contains any of the given types
    """
    if inner_types := list(get_inner_types(field.annotation, include_none=False)):
        return any(inner_type in types for inner_type in inner_types)
    return False


def get_inner_types(
    annotation: Any,  # noqa: ANN401
    include_none: bool = True,  # noqa: FBT001, FBT002
    unpack_list: bool = True,  # noqa: FBT001, FBT002
    unpack_literal: bool = True,  # noqa: FBT001, FBT002
) -> Generator[type, None, None]:
    """Recursively yield all inner types from a given type annotation.

    Args:
        annotation: The type annotation to process
        include_none: Whether to include NoneTypes in output
        unpack_list: Whether to unpack list types
        unpack_literal: Whether to unpack Literal types

    Returns:
        All inner types found within the annotation
    """
    # Check whether to unpack lists in addition to annotations and unions
    types_to_unpack = [Annotated, Union, UnionType] + ([list] if unpack_list else [])

    # Get the unsubscripted version of the given type annotation
    origin_type = get_origin(annotation)

    # If the origin should be unpacked
    if origin_type in types_to_unpack:
        for inner_type in get_args(annotation):
            # Recursively process each inner type
            yield from get_inner_types(
                inner_type, include_none, unpack_list, unpack_literal
            )

    # Handle Literal types based on the unpack_literal flag
    elif origin_type is Literal:
        if unpack_literal:
            yield origin_type  # Return Literal if unpacking is allowed
        else:
            yield annotation  # Return the full annotation if not

    # Yield the origin type if present
    elif origin_type is not None:
        yield origin_type

    # Yield the annotation if it is valid type, that isn't NoneType
    elif isinstance(annotation, type) and annotation is not NoneType:
        yield annotation

    # Optionally yield none types
    elif include_none and annotation in (None, NoneType):
        yield NoneType


@lru_cache(maxsize=4048)
def get_all_fields(model: type[BaseModel]) -> dict[str, GenericFieldInfo]:
    """Return a combined dict of defined and computed fields of a given model.

    This function combines both regular model fields and computed fields into a
    single dictionary using the GenericFieldInfo abstraction. Results are cached
    for performance.

    Args:
        model: The Pydantic model class to extract fields from.

    Returns:
        Dictionary mapping field names to GenericFieldInfo objects for all fields
        (both regular and computed) in the model.
    """
    return {
        **{
            name: GenericFieldInfo(
                alias=info.alias,
                annotation=info.annotation,
                frozen=bool(info.frozen),
            )
            for name, info in model.model_fields.items()
        },
        **{
            name: GenericFieldInfo(
                alias=info.alias,
                annotation=info.return_type,
                frozen=True,
            )
            for name, info in model.model_computed_fields.items()
        },
    }


@lru_cache(maxsize=4048)
def get_alias_lookup(model: type[BaseModel]) -> dict[str, str]:
    """Build a cached mapping from field alias to field names.

    Creates a dictionary that maps field aliases (or field names if no alias exists)
    back to the actual field names. This is useful for resolving field references
    when working with serialized data that may use aliases.

    Args:
        model: The Pydantic model class to build the alias lookup for.

    Returns:
        Dictionary mapping field aliases (or names) to actual field names.
    """
    return {
        field_info.alias or field_name: field_name
        for field_name, field_info in get_all_fields(model).items()
    }


@lru_cache(maxsize=4048)
def get_list_field_names(model: type[BaseModel]) -> list[str]:
    """Build a cached list of fields that look like lists.

    Analyzes the model's field annotations to identify fields that are list types.
    This includes direct list annotations and list types within unions.

    Args:
        model: The Pydantic model class to analyze.

    Returns:
        List of field names that have list-like type annotations.
    """
    field_names = []
    for field_name, field_info in get_all_fields(model).items():
        field_types = get_inner_types(field_info.annotation, unpack_list=False)
        if any(
            isinstance(field_type, type) and issubclass(field_type, list)
            for field_type in field_types
        ):
            field_names.append(field_name)
    return field_names


@lru_cache(maxsize=4048)
def get_field_names_allowing_none(model: type[BaseModel]) -> list[str]:
    """Build a cached list of fields that can be set to None.

    Tests each field's annotation by attempting to validate None against it.
    Fields that accept None without raising a ValidationError are considered
    nullable fields.

    Args:
        model: The Pydantic model class to analyze.

    Returns:
        List of field names that accept None as a valid value.
    """
    field_names: list[str] = []
    for field_name, field_info in get_all_fields(model).items():
        validator: TypeAdapter[Any] = TypeAdapter(field_info.annotation)
        try:
            validator.validate_python(None)
        except ValidationError:
            continue
        field_names.append(field_name)
    return field_names


def group_fields_by_class_name(
    model_classes_by_name: Mapping[str, type[BaseModel]],
    predicate: Callable[[GenericFieldInfo], bool],
) -> dict[str, list[str]]:
    """Group the field names by model class and filter them by the given predicate.

    For each model class, extracts all fields and applies the predicate function
    to filter them. Returns a mapping from class names to lists of field names
    that satisfy the predicate condition.

    Args:
        model_classes_by_name: Map from class names to model classes.
        predicate: Function to filter the fields of the classes by.

    Returns:
        Dictionary mapping class names to a list of field names filtered by `predicate`.
    """
    return {
        name: sorted(
            {
                field_name
                for field_name, field_info in get_all_fields(cls).items()
                if predicate(field_info)
            }
        )
        for name, cls in model_classes_by_name.items()
    }


def grouper(chunk_size: int, iterable: Iterable[T]) -> Iterator[Iterable[T | None]]:
    """Collect data into fixed-length chunks or blocks.

    Groups items from an iterable into fixed-size chunks. The last chunk may be
    padded with None values if the total number of items is not evenly divisible
    by the chunk size.

    Args:
        chunk_size: The size of each chunk.
        iterable: The iterable to group into chunks.

    Returns:
        Iterator of iterables, each containing chunk_size items (with None padding
        for the final chunk if necessary).
    """
    # https://docs.python.org/3.9/library/itertools.html#itertools-recipes
    args = [iter(iterable)] * chunk_size
    return zip_longest(*args, fillvalue=None)


def jitter_sleep(min_seconds: float, jitter_seconds: float) -> None:
    """Sleep a random amount of seconds within the given parameters.

    Args:
        min_seconds: The minimum time to sleep
        jitter_seconds: The variable sleep time added to the minimum
    """
    sleep(min_seconds + random() * jitter_seconds)  # noqa: S311


def ensure_list(values: list[T] | T | None) -> list[T]:
    """Put objects in lists, replace None with an empty list and return lists as is."""
    if values is None:
        return []
    if isinstance(values, list):
        return values
    return [values]


def deprecated(old_name: str, new_func: Callable[P, R]) -> Callable[P, R]:
    """Create a deprecated wrapper for a function."""

    @functools.wraps(new_func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        warnings.warn(
            f"{old_name} is deprecated, use {new_func.__name__} instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return new_func(*args, **kwargs)

    return wrapper



import string
from typing import Any, Self
from uuid import UUID, uuid4

from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler, json_schema
from pydantic_core import core_schema

_ALPHABET = string.ascii_letters + string.digits
IDENTIFIER_PATTERN = r"^[a-zA-Z0-9]{14,22}$"


class Identifier(str):
    """Common identifier class based on UUID version 4."""

    __slots__ = ()

    @classmethod
    def generate(cls, seed: int | None = None) -> Self:
        """Generate a new identifier from a seed or random UUID version 4.

        Creates a short alphanumerical identifier using UUID version 4 as the
        source. If a seed is provided, the UUID will be deterministic; otherwise
        a random UUID is generated.

        Args:
            seed: Optional integer seed for deterministic identifier generation.
                  If None, generates a random identifier.

        Returns:
            New identifier instance with the generated value.
        """
        # Inspired by https://pypi.org/project/shortuuid
        output = ""
        alpha_len = len(_ALPHABET)
        if seed is None:
            number = uuid4().int
        else:
            number = UUID(int=seed, version=4).int
        while number:
            number, digit = divmod(number, alpha_len)
            output += _ALPHABET[digit]
        return cls(output[::-1])

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,  # noqa: ANN401
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """Modify the core schema to add the ID regex."""
        return core_schema.chain_schema(
            [
                core_schema.str_schema(pattern=IDENTIFIER_PATTERN),
                core_schema.no_info_plain_validator_function(cls),
            ],
            serialization=core_schema.to_string_ser_schema(when_used="unless-none"),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, core_schema_: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> json_schema.JsonSchemaValue:
        """Modify the json schema to add the class name as title."""
        json_schema_ = handler(core_schema_)
        json_schema_ = handler.resolve_ref_schema(json_schema_)
        json_schema_["title"] = cls.__name__
        json_schema_["pattern"] = IDENTIFIER_PATTERN
        return json_schema_

    def __repr__(self) -> str:
        """Overwrite the default representation."""
        return f'{self.__class__.__name__}("{self}")'


# We have technically-identical subclasses of identifier types (one per entity-type).
# This allows us to annotate which entity-types are allowed on reference fields.
# For example `contact: MergedPersonIdentifier | MergedOrganizationIdentifier` tells us
# that a contact for an item has to be either a merged person or merged organization.
# We cannot validate this using pydantic, because all identifiers have the same
# format. But it helps for documentation purposes, allows us to generate a more
# precise JSON schema definitions and to derive database queries from the models.


class ExtractedIdentifier(Identifier):
    """Base class for all extracted identifiers."""


class ExtractedAccessPlatformIdentifier(ExtractedIdentifier):
    """Identifier for extracted access platforms."""


class ExtractedActivityIdentifier(ExtractedIdentifier):
    """Identifier for extracted activities."""


class ExtractedBibliographicResourceIdentifier(ExtractedIdentifier):
    """Identifier for extracted bibliographic resources."""


class ExtractedConsentIdentifier(ExtractedIdentifier):
    """Identifier for extracted consents."""


class ExtractedContactPointIdentifier(ExtractedIdentifier):
    """Identifier for extracted contact points."""


class ExtractedDistributionIdentifier(ExtractedIdentifier):
    """Identifier for extracted distributions."""


class ExtractedOrganizationIdentifier(ExtractedIdentifier):
    """Identifier for extracted organizations."""


class ExtractedOrganizationalUnitIdentifier(ExtractedIdentifier):
    """Identifier for extracted organizational units."""


class ExtractedPersonIdentifier(ExtractedIdentifier):
    """Identifier for extracted persons."""


class ExtractedPrimarySourceIdentifier(ExtractedIdentifier):
    """Identifier for extracted primary sources."""


class ExtractedResourceIdentifier(ExtractedIdentifier):
    """Identifier for extracted resources."""


class ExtractedVariableIdentifier(ExtractedIdentifier):
    """Identifier for extracted variables."""


class ExtractedVariableGroupIdentifier(ExtractedIdentifier):
    """Identifier for extracted variable groups."""


class MergedIdentifier(Identifier):
    """Base class for all merged identifiers."""


class MergedAccessPlatformIdentifier(MergedIdentifier):
    """Identifier for merged access platforms."""


class MergedActivityIdentifier(MergedIdentifier):
    """Identifier for merged activities."""


class MergedBibliographicResourceIdentifier(MergedIdentifier):
    """Identifier for merged bibliographic resources."""


class MergedConsentIdentifier(MergedIdentifier):
    """Identifier for merged consents."""


class MergedContactPointIdentifier(MergedIdentifier):
    """Identifier for merged contact points."""


class MergedDistributionIdentifier(MergedIdentifier):
    """Identifier for merged distributions."""


class MergedOrganizationIdentifier(MergedIdentifier):
    """Identifier for merged organizations."""


class MergedOrganizationalUnitIdentifier(MergedIdentifier):
    """Identifier for merged organizational units."""


class MergedPersonIdentifier(MergedIdentifier):
    """Identifier for merged persons."""


class MergedPrimarySourceIdentifier(MergedIdentifier):
    """Identifier for merged primary sources."""


class MergedResourceIdentifier(MergedIdentifier):
    """Identifier for merged resources."""


class MergedVariableIdentifier(MergedIdentifier):
    """Identifier for merged variables."""


class MergedVariableGroupIdentifier(MergedIdentifier):
    """Identifier for merged variable groups."""



from collections.abc import Generator
from typing import TypeVar

from mex.common.exceptions import MExError
from mex.common.logging import logger
from mex.common.models import (
    ExtractedOrganization,
    ExtractedOrganizationalUnit,
    MergedOrganizationalUnit,
)
from mex.common.organigram.extract import extract_organigram_units
from mex.common.organigram.models import OrganigramUnit
from mex.common.organigram.transform import (
    transform_organigram_unit_to_extracted_organizational_unit,
)
from mex.common.types import (
    MergedOrganizationalUnitIdentifier,
    MergedPrimarySourceIdentifier,
)

_TOrganizationalUnit = TypeVar(
    "_TOrganizationalUnit",
    MergedOrganizationalUnit,
    ExtractedOrganizationalUnit,
    OrganigramUnit,
)


def get_extracted_organizational_unit_with_parents(
    name: str,
    primary_source_id: MergedPrimarySourceIdentifier,
    rki_organization: ExtractedOrganization,
) -> list[ExtractedOrganizationalUnit]:
    """Pick the unit with the given name and transform it along with its parents.

    Args:
        name: Name (`identifierInPrimarySource`) of the organigram unit
        primary_source_id: Extracted primary source for the organigram
        rki_organization: RKI organization to which the unit belongs

    Returns:
        ExtractedOrganizationalUnit with its parents
    """
    organigram_units = extract_organigram_units()

    def with_parents(n: str) -> Generator[OrganigramUnit, None, None]:
        for unit in organigram_units:
            if n == unit.identifier or n in (
                t.value
                for texts in (unit.name, unit.shortName, unit.alternativeName)
                for t in texts
            ):
                yield unit
                if unit.parentUnit:
                    yield from with_parents(unit.parentUnit)
                break
        else:
            msg = f"could not find unit for {name}"
            raise MExError(msg)

    organigram_unit_with_parents = list(with_parents(name))
    extracted_unit_by_id_in_primary_source: dict[str, ExtractedOrganizationalUnit] = {}
    parent_id_in_primary_source_by_id_in_primary_source: dict[str, str] = {}

    for unit in organigram_unit_with_parents:
        extracted_unit = transform_organigram_unit_to_extracted_organizational_unit(
            unit, primary_source_id, rki_organization
        )
        extracted_unit_by_id_in_primary_source[unit.identifier] = extracted_unit
        if parent_identifier_in_primary_source := unit.parentUnit:
            parent_id_in_primary_source_by_id_in_primary_source[unit.identifier] = (
                parent_identifier_in_primary_source
            )

    for extracted_unit in extracted_unit_by_id_in_primary_source.values():
        identifier_in_primary_source = extracted_unit.identifierInPrimarySource
        if (  # noqa: SIM102
            parent_identifier_in_primary_source
            := parent_id_in_primary_source_by_id_in_primary_source.get(
                identifier_in_primary_source
            )
        ):
            if parent_unit := extracted_unit_by_id_in_primary_source.get(
                parent_identifier_in_primary_source
            ):
                extracted_unit.parentUnit = MergedOrganizationalUnitIdentifier(
                    parent_unit.stableTargetId
                )
    logger.info(
        "transformed %s organizational units",
        len(extracted_unit_by_id_in_primary_source),
    )
    return list(extracted_unit_by_id_in_primary_source.values())


def build_child_map(units: list[_TOrganizationalUnit]) -> dict[str, list[str]]:
    """Builds a dictionary with all children per unit from a list of units.

    Args:
        units: list of organizational units (Extracted, Merged, or OrganigramUnit)

    Returns:
        dict[str, list[str]]: Dictionary with a list of all child unit ids by unit id
    """
    child_map: dict[str, list[str]] = {}
    for unit in units:
        if unit.parentUnit is not None:
            child_map.setdefault(str(unit.parentUnit), []).append(str(unit.identifier))
    return child_map


def _collect_descendants(
    child_map: dict[str, list[str]],
    unit_id: str,
    descendant_ids: set[str],
) -> None:
    """Recursion to collect all children and their children, etc., depth first.

    Args:
        child_map: Dictionary with all child units of a unit by unit id
        unit_id: Unit identifier
        descendant_ids: Set of all descendant units

    Returns:
        None, the set descendant_ids is mutated in-place
    """
    for child_id in child_map.get(unit_id, []):
        descendant_ids.add(child_id)
        _collect_descendants(child_map, child_id, descendant_ids)


def find_descendants(units: list[_TOrganizationalUnit], parent_id: str) -> list[str]:
    """Find ids of all descendant (great{n}/grand/child) units for any parent unit id.

    Args:
        units: list of organizational units (Extracted, Merged, or OrganigramUnit)
                in which to search for descendants
        parent_id: identifier of the parent unit

    Returns:
        list of all unique descendant organizational unit ids (children,
                grandchildren, ...), excluding the starting parent_id
    """
    child_map = build_child_map(units)
    descendant_ids: set[str] = set()

    _collect_descendants(child_map, str(parent_id), descendant_ids)
    return list(descendant_ids)



from enum import Enum

import pytest
from pytest import MonkeyPatch

from mex.common.identity.base import BaseProvider
from mex.common.identity.memory import MemoryIdentityProvider
from mex.common.identity.models import Identity
from mex.common.identity.registry import (
    _PROVIDER_REGISTRY,
    get_provider,
    register_provider,
)
from mex.common.settings import SETTINGS_STORE, BaseSettings
from mex.common.types import IdentityProvider


class DummyIdentityProvider(Enum):
    DUMMY = "dummy"


class DummyProvider(BaseProvider):
    def __init__(self) -> None:
        pass

    def assign(self, *_: str) -> Identity:  # pragma: no cover
        raise RuntimeError

    def fetch(self, **_: str | None) -> list[Identity]:  # pragma: no cover
        raise RuntimeError

    def close(self) -> None:
        pass


class DummySettings(BaseSettings):
    identity_provider: IdentityProvider | DummyIdentityProvider = (
        IdentityProvider.MEMORY  # type: ignore[assignment]
    )


def test_register_provider_error() -> None:
    with pytest.raises(RuntimeError, match="Already registered identity provider"):
        register_provider(IdentityProvider.MEMORY, MemoryIdentityProvider)


def test_register_provider() -> None:
    register_provider(DummyIdentityProvider.DUMMY, DummyProvider)  # type: ignore[arg-type]

    assert _PROVIDER_REGISTRY[DummyIdentityProvider.DUMMY] == DummyProvider  # type: ignore[index]


def test_get_provider_error(monkeypatch: MonkeyPatch) -> None:
    # first remove the default provider from the registry
    monkeypatch.delitem(_PROVIDER_REGISTRY, IdentityProvider.MEMORY)

    # then get an error when we try to get an instance
    with pytest.raises(RuntimeError, match="Identity provider not implemented"):
        get_provider()


def test_get_provider() -> None:
    SETTINGS_STORE.reset()
    settings = DummySettings.get()
    settings.identity_provider = DummyIdentityProvider.DUMMY

    register_provider(DummyIdentityProvider.DUMMY, DummyProvider)  # type: ignore[arg-type]

    provider = get_provider()
    assert isinstance(provider, DummyProvider)



from functools import lru_cache
from typing import Any

from mex.common.exceptions import EmptySearchResultError, FoundMoreThanOneError
from mex.common.models import PaginatedItemsContainer
from mex.common.orcid.connector import OrcidConnector
from mex.common.orcid.models import OrcidRecord


@lru_cache(maxsize=128)
def get_orcid_record_by_id(orcid_id: str) -> OrcidRecord:
    """Get a single orcid record by id.

    Args:
        orcid_id: Unique identifier in ORCID system.

    Returns:
        Orcid record of the single matching id.
    """
    connector = OrcidConnector.get()
    return connector.get_record_by_id(orcid_id)


def get_orcid_record_by_name(
    given_names: str | None = None,
    family_name: str | None = None,
    given_and_family_names: str | None = None,
    filters: dict[str, Any] | None = None,
) -> OrcidRecord:
    """Get OrcidRecord of a single person for the given filters.

    Args:
        given_names: Optional given name of a person.
        family_name: Optional surname of a person.
        given_and_family_names: Optional full name of a person.
        filters: Key-value pairs representing ORCID search filters.

    Raises:
        EmptySearchResultError
        FoundMoreThanOneError

    Returns:
        OrcidRecord of the matching person by name.
    """
    connector = OrcidConnector.get()
    orcid_response = connector.search_records_by_name(
        given_names=given_names,
        family_name=family_name,
        given_and_family_names=given_and_family_names,
        filters=filters,
        skip=0,
        limit=1,
    )
    if orcid_response.num_found == 0:
        msg = "Cannot find orcid person for filters."
        raise EmptySearchResultError(msg)
    if orcid_response.num_found > 1:
        msg = f"Found multiple orcid persons for filters: {orcid_response.num_found}"
        raise FoundMoreThanOneError(msg)
    return get_orcid_record_by_id(
        orcid_response.result[0].orcid_identifier.path,
    )


def search_records_by_name(  # noqa: PLR0913
    given_names: str | None = None,
    family_name: str | None = None,
    given_and_family_names: str | None = None,
    filters: dict[str, Any] | None = None,
    skip: int = 0,
    limit: int = 10,
) -> PaginatedItemsContainer[OrcidRecord]:
    """Get all Orcid records for the given filters.

    Args:
        given_names: Optional given name of a person.
        family_name: Optional surname of a person.
        given_and_family_names: Optional full name of a person.
        filters: Key-value pairs representing ORCID search filters.
        skip: How many items to skip for pagination.
        limit: How many items to return in one page.

    Returns:
        Paginated container of orcid records.
    """
    connector = OrcidConnector.get()
    response = connector.search_records_by_name(
        given_names=given_names,
        family_name=family_name,
        given_and_family_names=given_and_family_names,
        filters=filters,
        skip=skip,
        limit=limit,
    )
    items = [
        get_orcid_record_by_id(
            item.orcid_identifier.path,
        )
        for item in response.result
    ]
    return PaginatedItemsContainer[OrcidRecord](items=items, total=response.num_found)



from typing import Annotated, ClassVar, Literal

from pydantic import Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    AccessRestriction,
    ExtractedDistributionIdentifier,
    License,
    Link,
    MergedAccessPlatformIdentifier,
    MergedDistributionIdentifier,
    MergedPrimarySourceIdentifier,
    MIMEType,
    Text,
    Year,
    YearMonth,
    YearMonthDay,
    YearMonthDayTime,
)


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["Distribution"], Field(frozen=True)]] = (
        "Distribution"
    )


class _OptionalLists(_Stem):
    accessURL: Annotated[
        list[Link],
        Field(json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#accessURL"]}),
    ] = []
    downloadURL: Annotated[
        list[Link],
        Field(json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#downloadURL"]}),
    ] = []


class _RequiredLists(_Stem):
    title: Annotated[
        list[Text],
        Field(
            min_length=1,
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/title"]},
        ),
    ]


class _SparseLists(_Stem):
    title: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/title"]}),
    ] = []


class _OptionalValues(_Stem):
    accessService: Annotated[
        MergedAccessPlatformIdentifier | None,
        Field(
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#accessService"]}
        ),
    ] = None
    license: Annotated[
        License | None,
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/license"]}),
    ] = None
    mediaType: Annotated[
        MIMEType | None,
        Field(
            json_schema_extra={
                "sameAs": [
                    "http://www.w3.org/ns/dcat#mediaType",
                    "http://purl.org/dc/terms/format",
                ]
            }
        ),
    ] = None
    modified: Annotated[
        YearMonthDayTime | YearMonthDay | YearMonth | Year | None,
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/modified"]}),
    ] = None


class _RequiredValues(_Stem):
    accessRestriction: Annotated[
        AccessRestriction,
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/accessRights"]}),
    ]
    issued: Annotated[
        YearMonthDayTime | YearMonthDay | YearMonth | Year,
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/issued"]}),
    ]


class _SparseValues(_Stem):
    accessRestriction: Annotated[
        AccessRestriction | None,
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/accessRights"]}),
    ] = None
    issued: Annotated[
        YearMonthDayTime | YearMonthDay | YearMonth | Year | None,
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/issued"]}),
    ] = None


class _VariadicValues(_Stem):
    accessRestriction: Annotated[
        list[AccessRestriction],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/accessRights"]}),
    ] = []
    accessService: Annotated[
        list[MergedAccessPlatformIdentifier],
        Field(
            json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#accessService"]}
        ),
    ] = []
    issued: Annotated[
        list[YearMonthDayTime | YearMonthDay | YearMonth | Year],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/issued"]}),
    ] = []
    license: Annotated[
        list[License],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/license"]}),
    ] = []
    mediaType: Annotated[
        list[MIMEType],
        Field(
            json_schema_extra={
                "sameAs": [
                    "http://www.w3.org/ns/dcat#mediaType",
                    "http://purl.org/dc/terms/format",
                ]
            }
        ),
    ] = []
    modified: Annotated[
        list[YearMonthDayTime | YearMonthDay | YearMonth | Year],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/modified"]}),
    ] = []


class BaseDistribution(
    _OptionalLists,
    _RequiredLists,
    _OptionalValues,
    _RequiredValues,
    json_schema_extra={
        "description": (
            "A specific representation of a dataset. A dataset might be available in "
            "multiple serializations that may differ in various ways, including "
            "natural language, media-type or format, schematic organization, temporal "
            "and spatial resolution, level of detail or profiles (which might specify "
            "any or all of the above) ([DCAT, 2020-02-04]"
            "(https://www.w3.org/TR/2020/REC-vocab-dcat-2-20200204/))."
        ),
        "sameAs": ["http://www.w3.org/ns/dcat#Distribution"],
        "title": "Distribution",
    },
):
    """All fields for a valid distribution except for provenance."""


class ExtractedDistribution(BaseDistribution, ExtractedData):
    """An automatically extracted metadata set describing a distribution."""

    entityType: Annotated[
        Literal["ExtractedDistribution"], Field(alias="$type", frozen=True)
    ] = "ExtractedDistribution"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(
        self,
    ) -> Annotated[
        ExtractedDistributionIdentifier,
        Field(
            json_schema_extra={"sameAs": ["http://purl.org/dc/elements/1.1/identifier"]}
        ),
    ]:
        """Return the computed identifier for this extracted item."""
        return self._get_identifier(ExtractedDistributionIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(self) -> MergedDistributionIdentifier:  # noqa: N802
        """Return the computed stableTargetId for this extracted item."""
        return self._get_stable_target_id(MergedDistributionIdentifier)


class MergedDistribution(BaseDistribution, MergedItem):
    """The result of merging all extracted items and rules for a distribution."""

    entityType: Annotated[
        Literal["MergedDistribution"], Field(alias="$type", frozen=True)
    ] = "MergedDistribution"
    identifier: Annotated[MergedDistributionIdentifier, Field(frozen=True)]


class PreviewDistribution(
    _OptionalLists, _SparseLists, _OptionalValues, _SparseValues, PreviewItem
):
    """Preview for merging all extracted items and rules for a distribution."""

    entityType: Annotated[
        Literal["PreviewDistribution"], Field(alias="$type", frozen=True)
    ] = "PreviewDistribution"
    identifier: Annotated[MergedDistributionIdentifier, Field(frozen=True)]


class AdditiveDistribution(
    _OptionalLists, _SparseLists, _OptionalValues, _SparseValues, AdditiveRule
):
    """Rule to add values to merged distribution items."""

    entityType: Annotated[
        Literal["AdditiveDistribution"], Field(alias="$type", frozen=True)
    ] = "AdditiveDistribution"


class SubtractiveDistribution(
    _OptionalLists, _SparseLists, _VariadicValues, SubtractiveRule
):
    """Rule to subtract values from merged distribution items."""

    entityType: Annotated[
        Literal["SubtractiveDistribution"], Field(alias="$type", frozen=True)
    ] = "SubtractiveDistribution"


class PreventiveDistribution(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged distribution items."""

    entityType: Annotated[
        Literal["PreventiveDistribution"], Field(alias="$type", frozen=True)
    ] = "PreventiveDistribution"
    accessRestriction: list[MergedPrimarySourceIdentifier] = []
    accessService: list[MergedPrimarySourceIdentifier] = []
    accessURL: list[MergedPrimarySourceIdentifier] = []
    downloadURL: list[MergedPrimarySourceIdentifier] = []
    issued: list[MergedPrimarySourceIdentifier] = []
    license: list[MergedPrimarySourceIdentifier] = []
    mediaType: list[MergedPrimarySourceIdentifier] = []
    modified: list[MergedPrimarySourceIdentifier] = []
    title: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    additive: AdditiveDistribution = AdditiveDistribution()
    subtractive: SubtractiveDistribution = SubtractiveDistribution()
    preventive: PreventiveDistribution = PreventiveDistribution()


class DistributionRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update a distribution item."""

    entityType: Annotated[
        Literal["DistributionRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "DistributionRuleSetRequest"


class DistributionRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve a distribution item."""

    entityType: Annotated[
        Literal["DistributionRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "DistributionRuleSetResponse"
    stableTargetId: MergedAccessPlatformIdentifier


class DistributionMapping(_Stem, BaseMapping):
    """Mapping for describing a distribution transformation."""

    entityType: Annotated[
        Literal["DistributionMapping"], Field(alias="$type", frozen=True)
    ] = "DistributionMapping"
    hadPrimarySource: Annotated[
        list[MappingField[MergedPrimarySourceIdentifier]], Field(min_length=1)
    ]
    identifierInPrimarySource: Annotated[list[MappingField[str]], Field(min_length=1)]
    accessRestriction: Annotated[
        list[MappingField[AccessRestriction]], Field(min_length=1)
    ]
    issued: Annotated[
        list[MappingField[YearMonthDayTime | YearMonthDay | YearMonth | Year]],
        Field(min_length=1),
    ]
    accessService: list[MappingField[MergedAccessPlatformIdentifier | None]] = []
    license: list[MappingField[License | None]] = []
    mediaType: list[MappingField[MIMEType | None]] = []
    modified: list[
        MappingField[YearMonthDayTime | YearMonthDay | YearMonth | Year | None]
    ] = []
    title: Annotated[list[MappingField[list[Text]]], Field(min_length=1)]
    accessURL: list[MappingField[list[Link]]] = []
    downloadURL: list[MappingField[list[Link]]] = []


class DistributionFilter(_Stem, BaseFilter):
    """Class for defining filter rules for distribution items."""

    entityType: Annotated[
        Literal["DistributionFilter"], Field(alias="$type", frozen=True)
    ] = "DistributionFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



import pytest
from pydantic import BaseModel, ValidationError

from mex.common.types import Link, LinkLanguage


class DummyModel(BaseModel):
    link: Link


def test_link_validation() -> None:
    with pytest.raises(ValidationError, match="Allowed input types are dict and str"):
        _ = DummyModel.model_validate({"link": 1})

    model = DummyModel.model_validate({"link": "https://example.com"})
    assert model.model_dump() == {
        "link": {
            "language": None,
            "title": None,
            "url": "https://example.com",
        }
    }

    model = DummyModel.model_validate(
        {"link": {"url": "https://example.com", "title": "Example", "language": "en"}}
    )
    assert model.model_dump() == {
        "link": {
            "language": LinkLanguage.EN,
            "title": "Example",
            "url": "https://example.com",
        }
    }

    # FR
    model = DummyModel.model_validate(
        {"link": {"url": "https://example.com", "title": "Exemple", "language": "fr"}}
    )
    assert model.model_dump() == {
        "link": {
            "language": LinkLanguage.FR,
            "title": "Exemple",
            "url": "https://example.com",
        }
    }

    # ES
    model = DummyModel.model_validate(
        {"link": {"url": "https://example.com", "title": "Ejemplo", "language": "es"}}
    )
    assert model.model_dump() == {
        "link": {
            "language": LinkLanguage.ES,
            "title": "Ejemplo",
            "url": "https://example.com",
        }
    }

    # RU
    model = DummyModel.model_validate(
        {"link": {"url": "https://example.com", "title": "Пример", "language": "ru"}}
    )
    assert model.model_dump() == {
        "link": {
            "language": LinkLanguage.RU,
            "title": "Пример",
            "url": "https://example.com",
        }
    }


def test_link_hash() -> None:
    link = Link(url="https://foo.bar", title="Hallo Welt.", language=LinkLanguage.DE)
    assert hash(link) == hash(("https://foo.bar", "Hallo Welt.", LinkLanguage.DE))



import json
from typing import Any

import pytest
from pytest import MonkeyPatch

from mex.common.wikidata.connector import WikidataAPIConnector
from mex.common.wikidata.extract import get_wikidata_organization
from tests.wikidata.conftest import TESTDATA_DIR


@pytest.mark.integration
def test_get_wikidata_organization() -> None:
    expected = {
        "identifier": "Q679041",
        "labels": {
            "de": {"language": "de", "value": "Robert Koch-Institut"},
            "en": {"language": "en", "value": "Robert Koch Institute"},
            "multiple": {"language": "mul", "value": "Robert Koch-Institut"},
        },
        "claims": {
            "website": [
                {"mainsnak": {"datavalue": {"value": {"text": "https://www.rki.de/"}}}},
                {
                    "mainsnak": {
                        "datavalue": {
                            "value": {
                                "text": "https://www.rki.de/DE/Home/homepage_node.html"
                            }
                        }
                    }
                },
                {
                    "mainsnak": {
                        "datavalue": {
                            "value": {
                                "text": "https://www.rki.de/EN/Home/homepage_node.html"
                            }
                        }
                    }
                },
            ],
            "isni_id": [
                {"mainsnak": {"datavalue": {"value": {"text": "0000000109403744"}}}}
            ],
            "ror_id": [{"mainsnak": {"datavalue": {"value": {"text": "01k5qnb77"}}}}],
            "official_name": [
                {
                    "mainsnak": {
                        "datavalue": {
                            "value": {"text": "Robert-Koch-Institut", "language": "de"}
                        }
                    }
                }
            ],
            "short_name": [
                {
                    "mainsnak": {
                        "datavalue": {"value": {"text": "RKI", "language": "en"}}
                    }
                },
                {
                    "mainsnak": {
                        "datavalue": {"value": {"text": "RKI", "language": "de"}}
                    }
                },
                {
                    "mainsnak": {
                        "datavalue": {"value": {"text": "IRK", "language": "fr"}}
                    }
                },
            ],
            "native_label": [
                {
                    "mainsnak": {
                        "datavalue": {
                            "value": {"text": "Robert Koch-Institut", "language": "de"}
                        }
                    }
                }
            ],
            "gepris_id": [{"mainsnak": {"datavalue": {"value": {"text": "10179"}}}}],
            "gnd_id": [{"mainsnak": {"datavalue": {"value": {"text": "17690-4"}}}}],
            "viaf_id": [{"mainsnak": {"datavalue": {"value": {"text": "123556639"}}}}],
        },
        "aliases": {
            "de": [
                {"language": "de", "value": "Robert Koch Institut"},
                {"language": "de", "value": "RKI"},
                {"language": "de", "value": "Robert-Koch-Institut"},
                {"language": "de", "value": "Kochsches Institut"},
            ],
            "en": [
                {"language": "en", "value": "RKI"},
                {"language": "en", "value": "Robert Koch-Institut"},
                {"language": "en", "value": "Robert Koch Institut"},
                {"language": "en", "value": "Robert-Koch-Institut"},
                {
                    "language": "en",
                    "value": "Reich Institution for the Combating of Contagious Diseases",
                },
            ],
        },
    }

    organization_details = get_wikidata_organization("Q679041")

    assert organization_details.model_dump(exclude_none=True) == expected


@pytest.mark.usefixtures("mocked_session_wikidata_api")
def test_get_wikidata_organization_mocked(monkeypatch: MonkeyPatch) -> None:
    expected = {
        "identifier": "Q26678",
        "labels": {
            "de": {"language": "de", "value": "TEST"},
            "en": {"language": "en", "value": "TEST"},
            "multiple": None,
        },
        "claims": {
            "website": [
                {
                    "mainsnak": {
                        "datavalue": {
                            "value": {
                                "text": "https://www.testgroup.com/",
                                "language": None,
                            }
                        }
                    }
                },
                {
                    "mainsnak": {
                        "datavalue": {
                            "value": {"text": "https://www.test.com", "language": None}
                        }
                    }
                },
            ],
            "isni_id": [
                {
                    "mainsnak": {
                        "datavalue": {
                            "value": {"text": "0000 0001 2308 257X", "language": None}
                        }
                    }
                }
            ],
            "ror_id": [
                {
                    "mainsnak": {
                        "datavalue": {"value": {"text": "05vs9tj88", "language": None}}
                    }
                },
                {
                    "mainsnak": {
                        "datavalue": {"value": {"text": "044kkbh92", "language": None}}
                    }
                },
            ],
            "official_name": [
                {
                    "mainsnak": {
                        "datavalue": {
                            "value": {
                                "text": "Bayerische Motoren Werke AG",
                                "language": "de",
                            }
                        }
                    }
                }
            ],
            "short_name": [],
            "native_label": [
                {
                    "mainsnak": {
                        "datavalue": {
                            "value": {"text": "test_native_name", "language": "de"}
                        }
                    }
                }
            ],
            "gepris_id": [],
            "gnd_id": [
                {
                    "mainsnak": {
                        "datavalue": {"value": {"text": "2005475-0", "language": None}}
                    }
                }
            ],
            "viaf_id": [
                {
                    "mainsnak": {
                        "datavalue": {"value": {"text": "129013645", "language": None}}
                    }
                }
            ],
        },
        "aliases": {
            "de": [
                {"language": "de", "value": "alias_de_1"},
                {"language": "de", "value": "alias_de_2"},
                {"language": "de", "value": "alias_de_3"},
            ],
            "en": [
                {"language": "en", "value": "alias_en_1"},
                {"language": "en", "value": "alias_en_2"},
                {"language": "en", "value": "alias_en_3"},
                {"language": "en", "value": "alias_en_4"},
            ],
        },
    }

    def mocked_item_details_response() -> Any:  # noqa: ANN401
        with (TESTDATA_DIR / "items_details.json").open(encoding="utf-8") as fh:
            data = json.load(fh)
            return data[0]

    monkeypatch.setattr(
        WikidataAPIConnector,
        "get_wikidata_item_details_by_id",
        lambda self, _: mocked_item_details_response(),
    )

    organization_details = get_wikidata_organization("Q26678")

    assert organization_details.model_dump() == expected


def test_get_wikidata_organization_malformed() -> None:
    with pytest.raises(ValueError, match="malformed wikidata url: foobar"):
        get_wikidata_organization("foobar")



import json

import pytest
from pydantic import BaseModel, ValidationError
from pytest import MonkeyPatch

from mex.common.types import VocabularyEnum
from mex.model import VOCABULARY_JSON_BY_NAME
from tests.types.conftest import TESTDATA_DIR


@pytest.fixture
def use_dummy_vocabulary(monkeypatch: MonkeyPatch) -> None:
    with (TESTDATA_DIR / "dummy-vocabulary.json").open() as fh:
        dummy_vocabulary = json.load(fh)
    monkeypatch.setitem(VOCABULARY_JSON_BY_NAME, "dummy_vocabulary", dummy_vocabulary)


@pytest.mark.usefixtures("use_dummy_vocabulary")
def test_vocabulary_enum_model() -> None:
    class DummyEnum(VocabularyEnum):
        __vocabulary__ = "dummy-vocabulary"

    # check enum names are loaded correctly
    assert [c.name for c in DummyEnum] == ["PREF_EN_ONE", "PREF_EN_TWO"]

    # check enum values are loaded correctly
    assert [c.value for c in DummyEnum] == [
        "https://mex.rki.de/item/dummy-concept-1",
        "https://mex.rki.de/item/dummy-concept-2",
    ]

    # check enum instance representation
    assert repr(DummyEnum["PREF_EN_ONE"]) == 'DummyEnum["PREF_EN_ONE"]'

    class DummyModel(BaseModel):
        dummy: DummyEnum

    # check wrong type raises error
    with pytest.raises(ValidationError):
        DummyModel.model_validate({"dummy": object()})

    # check wrong value raises error
    with pytest.raises(ValidationError):
        DummyModel.model_validate(
            {"dummy": "https://mex.rki.de/item/not-a-valid-concept"}
        )

    # check parsing from string works
    model = DummyModel.model_validate(
        {"dummy": "https://mex.rki.de/item/dummy-concept-2"}
    )
    assert model.dummy == DummyEnum["PREF_EN_TWO"]


@pytest.mark.usefixtures("use_dummy_vocabulary")
def test_vocabulary_enum_schema() -> None:
    class DummyEnum(VocabularyEnum):
        __vocabulary__ = "dummy-vocabulary"

    class DummyModel(BaseModel):
        dummy: DummyEnum

    assert DummyModel.model_json_schema() == {
        "properties": {
            "dummy": {
                "examples": ["https://mex.rki.de/item/dummy-concept-1"],
                "pattern": "https://mex.rki.de/item/[a-z0-9-]+",
                "title": "Dummy",
                "type": "string",
                "useScheme": "https://mex.rki.de/item/dummy-vocabulary",
            }
        },
        "required": ["dummy"],
        "title": "DummyModel",
        "type": "object",
    }


@pytest.mark.usefixtures("use_dummy_vocabulary")
def test_vocabulary_enum_find() -> None:
    class DummyEnum(VocabularyEnum):
        __vocabulary__ = "dummy-vocabulary"

    assert len(DummyEnum.__concepts__) == 2
    assert DummyEnum.find("not-found") is None

    found_enum = DummyEnum.find("pref-de-one")
    assert found_enum is not None
    assert found_enum.value == "https://mex.rki.de/item/dummy-concept-1"



from uuid import UUID

import pytest
from pydantic import BaseModel, ValidationError
from pytest import MonkeyPatch

from mex.common.types import IDENTIFIER_PATTERN, Identifier


class DummyIdentifier(Identifier):
    pass


class DummyModel(BaseModel):
    id: DummyIdentifier


def test_identifier_validation() -> None:
    model = DummyModel.model_validate({"id": "bFQoRhcVH5DIfZ"})
    assert model.id == DummyIdentifier("bFQoRhcVH5DIfZ")

    model = DummyModel.model_validate({"id": DummyIdentifier("bFQoRhcVH5DIfZ")})
    assert model.id == DummyIdentifier("bFQoRhcVH5DIfZ")

    model = DummyModel(id=DummyIdentifier("bFQoRhcVH5DIfZ"))
    assert model.id == DummyIdentifier("bFQoRhcVH5DIfZ")

    with pytest.raises(ValidationError):
        DummyModel.model_validate({"id": "baaiaaaboi!!!"})

    with pytest.raises(ValidationError):
        DummyModel.model_validate({"id": object()})


def test_identifier_serialization() -> None:
    model = DummyModel(id=DummyIdentifier("bFQoRhcVH5DIfZ"))
    raw = model.model_dump()

    assert raw == {"id": "bFQoRhcVH5DIfZ"}


def test_identifier_schema() -> None:
    assert DummyModel.model_json_schema() == {
        "properties": {
            "id": {
                "pattern": IDENTIFIER_PATTERN,
                "title": "DummyIdentifier",
                "type": "string",
            }
        },
        "required": ["id"],
        "title": "DummyModel",
        "type": "object",
    }


def test_identifier_repr() -> None:
    assert repr(Identifier("baaiaaaaaaaboi")) == 'Identifier("baaiaaaaaaaboi")'


def test_identifier_generate(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr(
        "mex.common.types.identifier.uuid4", lambda: UUID(int=42, version=4)
    )

    id_from_seed = Identifier.generate(seed=42)
    random_id = Identifier.generate()
    assert id_from_seed == random_id == Identifier("bFQoRhcVH5DHU6")



from mex.common.models import (
    EXTRACTED_MODEL_CLASSES,
    FILTER_MODEL_CLASSES,
    FilterField,
    PersonFilter,
)


def test_all_filter_classes_are_defined() -> None:
    stem_types = sorted(c.stemType for c in EXTRACTED_MODEL_CLASSES)
    assert sorted(c.stemType for c in FILTER_MODEL_CLASSES) == stem_types


def test_all_filter_fields_are_defined() -> None:
    for filter_cls in FILTER_MODEL_CLASSES:
        field_defs = {
            field_name: (field_info.annotation, field_info.default)
            for field_name, field_info in filter_cls.model_fields.items()
            if field_name != "entityType"
        }
        assert field_defs == {"fields": (list[FilterField], [])}


def test_filter_model_schema() -> None:
    assert PersonFilter.model_json_schema() == {
        "$defs": {
            "FilterField": {
                "additionalProperties": False,
                "description": "Filter definition for one field in the primary source.",
                "properties": {
                    "fieldInPrimarySource": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "fieldInPrimarySource",
                    },
                    "locationInPrimarySource": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "locationInPrimarySource",
                    },
                    "examplesInPrimarySource": {
                        "anyOf": [
                            {"items": {"type": "string"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "examplesInPrimarySource",
                    },
                    "filterRules": {
                        "items": {"$ref": "#/$defs/FilterRule"},
                        "minItems": 1,
                        "title": "filterRules",
                        "type": "array",
                    },
                    "comment": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "comment",
                    },
                },
                "required": ["filterRules"],
                "title": "FilterField",
                "type": "object",
            },
            "FilterRule": {
                "additionalProperties": False,
                "description": "A single filter rule to apply.",
                "properties": {
                    "forValues": {
                        "anyOf": [
                            {"items": {"type": "string"}, "type": "array"},
                            {"type": "null"},
                        ],
                        "default": None,
                        "title": "forValues",
                    },
                    "rule": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "rule",
                    },
                },
                "title": "FilterRule",
                "type": "object",
            },
        },
        "additionalProperties": False,
        "description": "Class for defining filter rules for person items.",
        "properties": {
            "$type": {
                "const": "PersonFilter",
                "default": "PersonFilter",
                "title": "$Type",
                "type": "string",
            },
            "fields": {
                "default": [],
                "items": {"$ref": "#/$defs/FilterField"},
                "title": "fields",
                "type": "array",
            },
        },
        "title": "PersonFilter",
        "type": "object",
    }



from typing import Annotated, ClassVar, Literal

from pydantic import Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    AccessRestriction,
    BibliographicResourceType,
    ExtractedBibliographicResourceIdentifier,
    Language,
    License,
    Link,
    MergedBibliographicResourceIdentifier,
    MergedDistributionIdentifier,
    MergedOrganizationalUnitIdentifier,
    MergedOrganizationIdentifier,
    MergedPersonIdentifier,
    MergedPrimarySourceIdentifier,
    Text,
    Year,
    YearMonth,
    YearMonthDay,
    YearMonthDayTime,
)

DoiStr = Annotated[
    str,
    Field(
        pattern="^https?://(?:dx\\.)?doi\\.org/[0-9]{2}\\.[0-9]{4,9}[-_.;()/:A-Za-z0-9]{0,256}$",
        examples=[
            "https://doi.org/10.1007/978-1-0716-2441-8_7",
            "https://doi.org/10.2807/1560-7917.ES.2022.27.46.2200849",
            "https://doi.org/10.3389/fmicb.2022.868887",
            "http://dx.doi.org/10.25646/5147",
            "https://doi.org/10.1016/j.vaccine.2022.11.065",
        ],
    ),
]
EditionStr = Annotated[
    str,
    Field(
        examples=[
            "5",
            "Band 2,1",
            "Band 2,2",
            "3rd edition",
        ],
    ),
]
IsbnIssnStr = Annotated[
    str,
    Field(
        examples=[
            "ISBN 90-70002-34-5",
            "ISBN 90-70002-34-x",
            "ISBN 90-70002-34-5x",
            "ISBN 90-700-02-34-5",
            "ISBN: 978-3-642-11746-6",
            "978-3-642-11746-6",
            "ISSN 0176-6996",
            "ISSN 1430-855X",
            "1430-8551",
            "1467-9442",
        ],
    ),
]
PagesStr = Annotated[
    str,
    Field(
        examples=[
            "1",
            "45-67",
            "45 - 67",
            "II",
            "XI",
            "10i",
        ]
    ),
]
PublicationPlaceStr = Annotated[
    str,
    Field(
        examples=[
            "Berlin",
            "Chigago",
            "NYC/NY",
            "Tampa, FL",
        ],
    ),
]
SectionStr = Annotated[
    str,
    Field(
        examples=[
            "Kapitel 1",
            "A Section About Public Health",
            "Chapter XII: The History of Public Health",
            "12",
            "A",
            "B.",
        ]
    ),
]
VolumeOrIssueStr = Annotated[
    str,
    Field(
        examples=[
            "2",
            "Q3",
            "11/12",
            "Winter '23",
        ]
    ),
]


class _Stem(BaseModel):
    stemType: ClassVar[
        Annotated[Literal["BibliographicResource"], Field(frozen=True)]
    ] = "BibliographicResource"


class _OptionalLists(_Stem):
    abstract: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/abstract"]}),
    ] = []
    alternateIdentifier: Annotated[
        list[str],
        Field(
            json_schema_extra={
                "sameAs": ["http://datacite.org/schema/kernel-4/alternateIdentifier"]
            }
        ),
    ] = []
    alternativeTitle: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/alternative"]}),
    ] = []
    bibliographicResourceType: Annotated[
        list[BibliographicResourceType],
        Field(json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/type"]}),
    ] = []
    contributingUnit: Annotated[
        list[MergedOrganizationalUnitIdentifier],
        Field(
            json_schema_extra={
                "subPropertyOf": ["http://purl.org/dc/terms/contributor"]
            }
        ),
    ] = []
    distribution: list[MergedDistributionIdentifier] = []
    editor: Annotated[
        list[MergedPersonIdentifier],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/contributor"]}),
    ] = []
    editorOfSeries: Annotated[
        list[MergedPersonIdentifier],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/contributor"]}),
    ] = []
    isbnIssn: Annotated[
        list[IsbnIssnStr],
        Field(
            json_schema_extra={
                "sameAs": ["http://datacite.org/schema/kernel-4/alternateIdentifier"]
            }
        ),
    ] = []
    journal: list[Text] = []
    keyword: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#keyword"]}),
    ] = []
    language: Annotated[
        list[Language],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/language"]}),
    ] = []
    publisher: Annotated[
        list[MergedOrganizationIdentifier],
        Field(json_schema_extra={"sameAs": "http://purl.org/dc/terms/publisher"}),
    ] = []
    repositoryURL: list[Link] = []
    subtitle: Annotated[
        list[Text],
        Field(json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/title"]}),
    ] = []
    titleOfBook: Annotated[
        list[Text],
        Field(json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/title"]}),
    ] = []
    titleOfSeries: Annotated[
        list[Text],
        Field(json_schema_extra={"subPropertyOf": ["http://purl.org/dc/terms/title"]}),
    ] = []


class _RequiredLists(_Stem):
    creator: Annotated[
        list[MergedPersonIdentifier],
        Field(
            min_length=1,
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/creator"]},
        ),
    ]
    title: Annotated[
        list[Text],
        Field(
            min_length=1,
            json_schema_extra={"sameAs": ["http://purl.org/dc/terms/title"]},
        ),
    ]


class _SparseLists(_Stem):
    creator: list[MergedPersonIdentifier] = []
    title: list[Text] = []


class _OptionalValues(_Stem):
    doi: DoiStr | None = None
    edition: EditionStr | None = None
    issue: VolumeOrIssueStr | None = None
    issued: Annotated[
        YearMonthDayTime | YearMonthDay | YearMonth | Year | None,
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/created"]}),
    ] = None
    license: Annotated[
        License | None,
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/license"]}),
    ] = None
    pages: PagesStr | None = None
    publicationPlace: PublicationPlaceStr | None = None
    publicationYear: Annotated[
        Year | None,
        Field(
            json_schema_extra={
                "sameAs": "http://datacite.org/schema/kernel-4/publicationYear"
            }
        ),
    ] = None
    section: SectionStr | None = None
    volume: VolumeOrIssueStr | None = None
    volumeOfSeries: VolumeOrIssueStr | None = None


class _RequiredValues(_Stem):
    accessRestriction: Annotated[
        AccessRestriction,
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/accessRights"]}),
    ]


class _SparseValues(_Stem):
    accessRestriction: AccessRestriction | None = None


class _VariadicValues(_Stem):
    accessRestriction: list[AccessRestriction] = []
    doi: list[DoiStr] = []
    edition: list[EditionStr] = []
    issue: list[VolumeOrIssueStr] = []
    issued: list[YearMonthDayTime | YearMonthDay | YearMonth | Year] = []
    license: list[License] = []
    pages: list[PagesStr] = []
    publicationPlace: list[PublicationPlaceStr] = []
    publicationYear: list[Year] = []
    repositoryURL: list[Link] = []
    section: list[SectionStr] = []
    volume: list[VolumeOrIssueStr] = []
    volumeOfSeries: list[VolumeOrIssueStr] = []


class BaseBibliographicResource(
    _OptionalLists,
    _RequiredLists,
    _OptionalValues,
    _RequiredValues,
    json_schema_extra={
        "description": "A book, article, or other documentary resource.",
        "sameAs": ["http://purl.org/dc/terms/BibliographicResource"],
        "title": "Bibliographic Resource",
    },
):
    """All fields for a valid bibliographic resource except for provenance."""


class ExtractedBibliographicResource(BaseBibliographicResource, ExtractedData):
    """An automatically extracted metadata item describing a bibliographic resource."""

    entityType: Annotated[
        Literal["ExtractedBibliographicResource"], Field(alias="$type", frozen=True)
    ] = "ExtractedBibliographicResource"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(
        self,
    ) -> Annotated[
        ExtractedBibliographicResourceIdentifier,
        Field(
            json_schema_extra={"sameAs": ["http://purl.org/dc/elements/1.1/identifier"]}
        ),
    ]:
        """Return the computed identifier for this extracted item."""
        return self._get_identifier(ExtractedBibliographicResourceIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(self) -> MergedBibliographicResourceIdentifier:  # noqa: N802
        """Return the computed stableTargetId for this extracted item."""
        return self._get_stable_target_id(MergedBibliographicResourceIdentifier)


class MergedBibliographicResource(BaseBibliographicResource, MergedItem):
    """The result of merging all extracted items and rules for a bibliographic resource."""  # noqa: E501

    entityType: Annotated[
        Literal["MergedBibliographicResource"], Field(alias="$type", frozen=True)
    ] = "MergedBibliographicResource"
    identifier: Annotated[MergedBibliographicResourceIdentifier, Field(frozen=True)]


class PreviewBibliographicResource(
    _OptionalLists, _SparseLists, _OptionalValues, _SparseValues, PreviewItem
):
    """Preview for merging all extracted items and rules for a bibliographic resource."""  # noqa: E501

    entityType: Annotated[
        Literal["PreviewBibliographicResource"], Field(alias="$type", frozen=True)
    ] = "PreviewBibliographicResource"
    identifier: Annotated[MergedBibliographicResourceIdentifier, Field(frozen=True)]


class AdditiveBibliographicResource(
    _OptionalLists, _SparseLists, _OptionalValues, _SparseValues, AdditiveRule
):
    """Rule to add values to merged bibliographic resource items."""

    entityType: Annotated[
        Literal["AdditiveBibliographicResource"], Field(alias="$type", frozen=True)
    ] = "AdditiveBibliographicResource"


class SubtractiveBibliographicResource(
    _OptionalLists, _SparseLists, _VariadicValues, SubtractiveRule
):
    """Rule to subtract values from merged bibliographic resource items."""

    entityType: Annotated[
        Literal["SubtractiveBibliographicResource"], Field(alias="$type", frozen=True)
    ] = "SubtractiveBibliographicResource"


class PreventiveBibliographicResource(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged bibliographic resource items."""  # noqa: E501

    entityType: Annotated[
        Literal["PreventiveBibliographicResource"], Field(alias="$type", frozen=True)
    ] = "PreventiveBibliographicResource"
    abstract: list[MergedPrimarySourceIdentifier] = []
    accessRestriction: list[MergedPrimarySourceIdentifier] = []
    alternateIdentifier: list[MergedPrimarySourceIdentifier] = []
    alternativeTitle: list[MergedPrimarySourceIdentifier] = []
    bibliographicResourceType: list[MergedPrimarySourceIdentifier] = []
    contributingUnit: list[MergedPrimarySourceIdentifier] = []
    creator: list[MergedPrimarySourceIdentifier] = []
    distribution: list[MergedPrimarySourceIdentifier] = []
    doi: list[MergedPrimarySourceIdentifier] = []
    edition: list[MergedPrimarySourceIdentifier] = []
    editor: list[MergedPrimarySourceIdentifier] = []
    editorOfSeries: list[MergedPrimarySourceIdentifier] = []
    isbnIssn: list[MergedPrimarySourceIdentifier] = []
    issue: list[MergedPrimarySourceIdentifier] = []
    issued: list[MergedPrimarySourceIdentifier] = []
    journal: list[MergedPrimarySourceIdentifier] = []
    keyword: list[MergedPrimarySourceIdentifier] = []
    language: list[MergedPrimarySourceIdentifier] = []
    license: list[MergedPrimarySourceIdentifier] = []
    pages: list[MergedPrimarySourceIdentifier] = []
    publicationPlace: list[MergedPrimarySourceIdentifier] = []
    publicationYear: list[MergedPrimarySourceIdentifier] = []
    publisher: list[MergedPrimarySourceIdentifier] = []
    repositoryURL: list[MergedPrimarySourceIdentifier] = []
    section: list[MergedPrimarySourceIdentifier] = []
    subtitle: list[MergedPrimarySourceIdentifier] = []
    title: list[MergedPrimarySourceIdentifier] = []
    titleOfBook: list[MergedPrimarySourceIdentifier] = []
    titleOfSeries: list[MergedPrimarySourceIdentifier] = []
    volume: list[MergedPrimarySourceIdentifier] = []
    volumeOfSeries: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    additive: AdditiveBibliographicResource = AdditiveBibliographicResource()
    subtractive: SubtractiveBibliographicResource = SubtractiveBibliographicResource()
    preventive: PreventiveBibliographicResource = PreventiveBibliographicResource()


class BibliographicResourceRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update a bibliographic resource item."""

    entityType: Annotated[
        Literal["BibliographicResourceRuleSetRequest"],
        Field(alias="$type", frozen=True),
    ] = "BibliographicResourceRuleSetRequest"


class BibliographicResourceRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve a bibliographic resource item."""

    entityType: Annotated[
        Literal["BibliographicResourceRuleSetResponse"],
        Field(alias="$type", frozen=True),
    ] = "BibliographicResourceRuleSetResponse"
    stableTargetId: MergedBibliographicResourceIdentifier


class BibliographicResourceMapping(_Stem, BaseMapping):
    """Mapping for describing a bibliographic resource transformation."""

    entityType: Annotated[
        Literal["BibliographicResourceMapping"], Field(alias="$type", frozen=True)
    ] = "BibliographicResourceMapping"
    hadPrimarySource: Annotated[
        list[MappingField[MergedPrimarySourceIdentifier]], Field(min_length=1)
    ]
    identifierInPrimarySource: Annotated[list[MappingField[str]], Field(min_length=1)]
    accessRestriction: Annotated[
        list[MappingField[AccessRestriction]], Field(min_length=1)
    ]
    doi: list[MappingField[DoiStr | None]] = []
    edition: list[MappingField[EditionStr | None]] = []
    issue: list[MappingField[VolumeOrIssueStr | None]] = []
    issued: list[
        MappingField[YearMonthDayTime | YearMonthDay | YearMonth | Year | None]
    ] = []
    license: list[MappingField[License | None]] = []
    pages: list[MappingField[PagesStr | None]] = []
    publicationPlace: list[MappingField[PublicationPlaceStr | None]] = []
    publicationYear: list[MappingField[Year | None]] = []
    repositoryURL: list[MappingField[Link | None]] = []
    section: list[MappingField[SectionStr | None]] = []
    volume: list[MappingField[VolumeOrIssueStr | None]] = []
    volumeOfSeries: list[MappingField[VolumeOrIssueStr | None]] = []
    creator: Annotated[
        list[MappingField[list[MergedPersonIdentifier]]], Field(min_length=1)
    ]
    title: Annotated[list[MappingField[list[Text]]], Field(min_length=1)]
    abstract: list[MappingField[list[Text]]] = []
    alternateIdentifier: list[MappingField[list[str]]] = []
    alternativeTitle: list[MappingField[list[Text]]] = []
    bibliographicResourceType: list[MappingField[list[BibliographicResourceType]]] = []
    contributingUnit: list[MappingField[list[MergedOrganizationalUnitIdentifier]]] = []
    distribution: list[MappingField[list[MergedDistributionIdentifier]]] = []
    editor: list[MappingField[list[MergedPersonIdentifier]]] = []
    editorOfSeries: list[MappingField[list[MergedPersonIdentifier]]] = []
    isbnIssn: list[MappingField[list[IsbnIssnStr]]] = []
    journal: list[MappingField[list[Text]]] = []
    keyword: list[MappingField[list[Text]]] = []
    language: list[MappingField[list[Language]]] = []
    publisher: list[MappingField[list[MergedOrganizationIdentifier]]] = []
    subtitle: list[MappingField[list[Text]]] = []
    titleOfBook: list[MappingField[list[Text]]] = []
    titleOfSeries: list[MappingField[list[Text]]] = []


class BibliographicResourceFilter(_Stem, BaseFilter):
    """Class for defining filter rules for bibliographic resource items."""

    entityType: Annotated[
        Literal["BibliographicResourceFilter"], Field(alias="$type", frozen=True)
    ] = "BibliographicResourceFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



from unittest.mock import MagicMock, Mock

import pytest
import requests

from mex.common.exceptions import EmptySearchResultError
from mex.common.wikidata.connector import WikidataAPIConnector


@pytest.mark.integration
def test_get_wikidata_item_details_by_id() -> None:
    """Test if items details can be fetched by its ID."""
    connector = WikidataAPIConnector.get()
    response = connector.get_wikidata_item_details_by_id("Q26678")

    assert list(response.keys()) == [
        "pageid",
        "ns",
        "title",
        "lastrevid",
        "modified",
        "type",
        "id",
        "labels",
        "descriptions",
        "aliases",
        "claims",
        "sitelinks",
    ]
    assert response.get("title") == "Q26678"
    assert response.get("type") == "item"


@pytest.mark.integration
def test_get_wikidata_item_details_by_id_error() -> None:
    """Test if items details can be fetched by its ID."""
    connector = WikidataAPIConnector.get()
    with pytest.raises(EmptySearchResultError, match="NOPE_NOPE_NOPE"):
        connector.get_wikidata_item_details_by_id("NOPE_NOPE_NOPE")


def test_get_wikidata_items_details_by_id_mocked(
    mocked_session_wikidata_api: MagicMock,
) -> None:
    expected = {
        "entities": {
            "Q26678": {
                "pageid": 30097,
                "ns": 0,
                "title": "Q26678",
                "lastrevid": 1817510361,
                "modified": "2023-01-23T09:14:17Z",
                "type": "item",
                "id": "Q26678",
                "descriptions": {
                    "en": {
                        "language": "en",
                        "value": "German automobile manufacturer, and conglomerate",
                    },
                    "de": {
                        "language": "de",
                        "value": "deutscher Automobil- und Motorradhersteller",
                    },
                },
            }
        },
        "success": 1,
    }

    mocked_session_wikidata_api.request = MagicMock(
        return_value=Mock(
            spec=requests.Response,
            json=MagicMock(return_value={"entities": {"Q26678": expected}}),
            status_code=200,
        )
    )

    connector = WikidataAPIConnector.get()

    response = connector.get_wikidata_item_details_by_id("Q26678")

    assert response == expected



from mex.common.connector import CONNECTOR_STORE, BaseConnector


class DummyConnector(BaseConnector):
    def __init__(self) -> None:
        self.closed = False

    def metrics(self) -> dict[str, int]:
        return {"foo": 42}

    def close(self) -> None:
        self.closed = True


def test_connector_get() -> None:
    assert len(list(CONNECTOR_STORE)) == 0

    connector = DummyConnector.get()
    assert isinstance(connector, DummyConnector)

    assert len(list(CONNECTOR_STORE)) == 1

    assert DummyConnector.get() is connector


def test_connector_store_reset() -> None:
    connector = DummyConnector.get()
    assert len(list(CONNECTOR_STORE)) == 1

    CONNECTOR_STORE.reset()

    assert connector.closed is True
    assert len(list(CONNECTOR_STORE)) == 0


def test_connector_store_metrics() -> None:
    DummyConnector.get()
    assert len(list(CONNECTOR_STORE)) == 1

    assert CONNECTOR_STORE.metrics() == {"dummy_connector_foo": 42}

    CONNECTOR_STORE.reset()
    assert CONNECTOR_STORE.metrics() == {}



import time
from typing import Any
from unittest.mock import MagicMock, Mock, call

import pytest
import requests
from pytest import LogCaptureFixture, MonkeyPatch
from requests import ConnectTimeout, JSONDecodeError, RequestException, Response, codes

from mex.common.connector import CONNECTOR_STORE, HTTPConnector
from mex.common.exceptions import TimedReadTimeout


class DummyHTTPConnector(HTTPConnector):
    TIMEOUT = 1
    TIMEOUT_MAX = 2
    PROPORTIONAL_BACKOFF_MIN = 0.1

    def _set_url(self) -> None:
        self.url = "https://www.example.com"

    def _check_availability(self) -> None:
        pass


@pytest.fixture
def mocked_dummy_session(monkeypatch: MonkeyPatch) -> MagicMock:
    """Mock the DummyHTTPConnector with a MagicMock session and return that."""
    mocked_session = MagicMock(spec=requests.Session, name="dummy_session")
    mocked_session.request = MagicMock(
        return_value=Mock(spec=requests.Response, status_code=200)
    )

    def set_mocked_session(self: DummyHTTPConnector) -> None:
        self.session = mocked_session

    monkeypatch.setattr(DummyHTTPConnector, "_set_session", set_mocked_session)
    return mocked_session


def test_init_mocked(mocked_dummy_session: MagicMock, monkeypatch: MonkeyPatch) -> None:
    def _check_availability(self: DummyHTTPConnector) -> None:
        self.request("GET", "_system/check")

    monkeypatch.setattr(DummyHTTPConnector, "_check_availability", _check_availability)

    connector = DummyHTTPConnector.get()
    connector.request("GET", "_system/check")
    assert connector.url == "https://www.example.com"
    assert mocked_dummy_session.request.call_args_list[-1] == call(
        "GET",
        "https://www.example.com/_system/check",
        None,
        timeout=DummyHTTPConnector.TIMEOUT,
        headers={
            "Accept": "application/json",
            "User-Agent": "rki/mex",
        },
    )


@pytest.mark.usefixtures("mocked_dummy_session")
def test_reset_all_connectors(mocked_dummy_session: MagicMock) -> None:
    DummyHTTPConnector.get()
    assert len(list(CONNECTOR_STORE)) == 1

    CONNECTOR_STORE.reset()
    mocked_dummy_session.close.assert_called_once_with()

    assert len(list(CONNECTOR_STORE)) == 0


@pytest.mark.parametrize(
    ("sent_payload", "mocked_response", "expected_response", "expected_kwargs"),
    [
        (
            {},
            MagicMock(status_code=204, json=MagicMock(side_effect=JSONDecodeError)),
            {},
            {
                "headers": {
                    "Accept": "application/json",
                    "User-Agent": "rki/mex",
                },
            },
        ),
        (
            {},
            MagicMock(status_code=200, json=MagicMock(return_value={"foo": "bar"})),
            {"foo": "bar"},
            {
                "headers": {
                    "Accept": "application/json",
                    "User-Agent": "rki/mex",
                },
            },
        ),
        (
            {"q": "SELECT status;"},
            MagicMock(status_code=200, json=MagicMock(return_value={"status": 42})),
            {"status": 42},
            {
                "headers": {
                    "Accept": "application/json",
                    "User-Agent": "rki/mex",
                },
                "data": '{"q": "SELECT status;"}',
            },
        ),
    ],
    ids=[
        "sending no payload and receiving 204 response",
        "sending no payload and receiving 200 response",
        "sending payload and receiving 200 response",
    ],
)
def test_request_success(
    monkeypatch: MonkeyPatch,
    sent_payload: dict[str, Any] | None,
    mocked_response: Response,
    expected_response: dict[str, Any],
    expected_kwargs: dict[str, Any],
) -> None:
    mocked_session = MagicMock(spec=requests.Session, name="dummy_session")
    mocked_session.request = MagicMock(return_value=mocked_response)

    def set_mocked_session(self: DummyHTTPConnector) -> None:
        self.session = mocked_session

    monkeypatch.setattr(DummyHTTPConnector, "_set_session", set_mocked_session)

    connector = DummyHTTPConnector.get()

    actual_response = connector.request("POST", "things", payload=sent_payload)

    assert actual_response == expected_response
    assert mocked_session.request.call_args_list[-1] == call(
        "POST",
        "https://www.example.com/things",
        None,
        timeout=DummyHTTPConnector.TIMEOUT,
        **expected_kwargs,
    )


@pytest.mark.parametrize(
    (
        "fake_response_time",
        "error_instance",
        "expected_response",
        "expected_retries",
    ),
    [
        (
            0.0,
            RequestException(
                response=Mock(
                    status_code=codes.internal_server_error,
                    json=MagicMock(
                        return_value={"status": codes.internal_server_error}
                    ),
                )
            ),
            "TimedServerError()",
            5,
        ),
        (
            0.0,
            RequestException(
                response=Mock(
                    status_code=codes.too_many_requests,
                    json=MagicMock(return_value={"status": codes.too_many_requests}),
                )
            ),
            "TimedTooManyRequests()",
            5,
        ),
        (
            0.0,
            ConnectTimeout("connect took too long"),
            "ConnectTimeout('connect took too long')",
            3,
        ),
        (
            1.5,
            TimedReadTimeout("read took too long", seconds=1.5),
            "TimedReadTimeout('read took too long')",
            5,
        ),
    ],
    ids=[
        "internal server error",
        "too many requests",
        "connect timeout",
        "timed read timeout",
    ],
)
def test_request_failure(  # noqa: PLR0913
    monkeypatch: MonkeyPatch,
    caplog: LogCaptureFixture,
    fake_response_time: float,
    error_instance: Exception,
    expected_response: str | dict[str, Any],
    expected_retries: int,
) -> None:
    def mock_request(*_: Any, **__: Any) -> Response:  # noqa: ANN401
        time.sleep(fake_response_time)
        raise error_instance

    mocked_session = MagicMock(spec=requests.Session, name="dummy_session")
    mocked_session.request = mock_request

    def set_mocked_session(self: DummyHTTPConnector) -> None:
        self.session = mocked_session

    monkeypatch.setattr(DummyHTTPConnector, "_set_session", set_mocked_session)

    connector = DummyHTTPConnector.get()

    try:
        response_or_error: Any = connector.request("POST", "things", payload=[])
    except RequestException as error:
        response_or_error = repr(error)

    assert response_or_error == expected_response
    assert len(caplog.messages) == expected_retries
    assert all(
        message.startswith("Backing off _send_request")
        for message in caplog.messages[: expected_retries - 1]
    )
    assert caplog.messages[-1].startswith(
        f"Giving up _send_request(...) after {expected_retries} tries"
    )



from typing import Annotated, ClassVar, Literal

from pydantic import AfterValidator, Field, computed_field

from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField
from mex.common.models.base.mapping import BaseMapping, MappingField
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.preview_item import PreviewItem
from mex.common.models.base.rules import (
    AdditiveRule,
    PreventiveRule,
    RuleSet,
    SubtractiveRule,
)
from mex.common.types import (
    ExtractedPrimarySourceIdentifier,
    Identifier,
    Link,
    MergedContactPointIdentifier,
    MergedOrganizationalUnitIdentifier,
    MergedPersonIdentifier,
    MergedPrimarySourceIdentifier,
    Text,
)

VersionStr = Annotated[
    str,
    Field(
        examples=["v1", "2023-01-16", "Schema 9"],
    ),
]
AnyContactIdentifier = Annotated[
    MergedOrganizationalUnitIdentifier
    | MergedPersonIdentifier
    | MergedContactPointIdentifier,
    AfterValidator(Identifier),
]


class _Stem(BaseModel):
    stemType: ClassVar[Annotated[Literal["PrimarySource"], Field(frozen=True)]] = (
        "PrimarySource"
    )


class _OptionalLists(_Stem):
    alternativeTitle: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/alternative"]}),
    ] = []
    contact: Annotated[
        list[AnyContactIdentifier],
        Field(json_schema_extra={"sameAs": ["http://www.w3.org/ns/dcat#contactPoint"]}),
    ] = []
    description: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/description"]}),
    ] = []
    documentation: Annotated[
        list[Link],
        Field(
            json_schema_extra={
                "subPropertyOf": ["http://purl.org/dc/terms/isReferencedBy"]
            }
        ),
    ] = []
    locatedAt: list[Link] = []
    title: Annotated[
        list[Text],
        Field(json_schema_extra={"sameAs": ["http://purl.org/dc/terms/title"]}),
    ] = []
    unitInCharge: Annotated[
        list[MergedOrganizationalUnitIdentifier],
        Field(
            json_schema_extra={"sameAs": ["http://dcat-ap.de/def/dcatde/maintainer"]}
        ),
    ] = []


class _OptionalValues(_Stem):
    version: VersionStr | None = None


class _VariadicValues(_Stem):
    version: list[VersionStr] = []


class BasePrimarySource(
    _OptionalLists,
    _OptionalValues,
    json_schema_extra={
        "description": (
            "A collection of information, that is managed and curated by an RKI unit "
            "and lists activities and/or resources."
        ),
        "sameAs": [
            "http://www.w3.org/ns/dcat#Catalog",
            "http://www.w3.org/ns/prov#PrimarySource",
        ],
        "title": "Primary Source",
    },
):
    """All fields for a valid primary source except for provenance."""


class ExtractedPrimarySource(BasePrimarySource, ExtractedData):
    """An automatically extracted metadata set describing a primary source."""

    entityType: Annotated[
        Literal["ExtractedPrimarySource"], Field(alias="$type", frozen=True)
    ] = "ExtractedPrimarySource"

    @computed_field  # type: ignore[prop-decorator]
    @property
    def identifier(
        self,
    ) -> Annotated[
        ExtractedPrimarySourceIdentifier,
        Field(
            json_schema_extra={"sameAs": ["http://purl.org/dc/elements/1.1/identifier"]}
        ),
    ]:
        """Return the computed identifier for this extracted item."""
        return self._get_identifier(ExtractedPrimarySourceIdentifier)

    @computed_field  # type: ignore[prop-decorator]
    @property
    def stableTargetId(self) -> MergedPrimarySourceIdentifier:  # noqa: N802
        """Return the computed stableTargetId for this extracted item."""
        return self._get_stable_target_id(MergedPrimarySourceIdentifier)


class MergedPrimarySource(BasePrimarySource, MergedItem):
    """The result of merging all extracted items and rules for a primary source."""

    entityType: Annotated[
        Literal["MergedPrimarySource"], Field(alias="$type", frozen=True)
    ] = "MergedPrimarySource"
    identifier: Annotated[MergedPrimarySourceIdentifier, Field(frozen=True)]


class PreviewPrimarySource(_OptionalLists, _OptionalValues, PreviewItem):
    """Preview for merging all extracted items and rules for a primary source."""

    entityType: Annotated[
        Literal["PreviewPrimarySource"], Field(alias="$type", frozen=True)
    ] = "PreviewPrimarySource"
    identifier: Annotated[MergedPrimarySourceIdentifier, Field(frozen=True)]


class AdditivePrimarySource(_OptionalLists, _OptionalValues, AdditiveRule):
    """Rule to add values to merged primary source items."""

    entityType: Annotated[
        Literal["AdditivePrimarySource"], Field(alias="$type", frozen=True)
    ] = "AdditivePrimarySource"


class SubtractivePrimarySource(_OptionalLists, _VariadicValues, SubtractiveRule):
    """Rule to subtract values from merged primary source items."""

    entityType: Annotated[
        Literal["SubtractivePrimarySource"], Field(alias="$type", frozen=True)
    ] = "SubtractivePrimarySource"


class PreventivePrimarySource(_Stem, PreventiveRule):
    """Rule to prevent primary sources for fields of merged primary source items."""

    entityType: Annotated[
        Literal["PreventivePrimarySource"], Field(alias="$type", frozen=True)
    ] = "PreventivePrimarySource"
    alternativeTitle: list[MergedPrimarySourceIdentifier] = []
    contact: list[MergedPrimarySourceIdentifier] = []
    description: list[MergedPrimarySourceIdentifier] = []
    documentation: list[MergedPrimarySourceIdentifier] = []
    locatedAt: list[MergedPrimarySourceIdentifier] = []
    title: list[MergedPrimarySourceIdentifier] = []
    unitInCharge: list[MergedPrimarySourceIdentifier] = []
    version: list[MergedPrimarySourceIdentifier] = []


class _BaseRuleSet(_Stem, RuleSet):
    additive: AdditivePrimarySource = AdditivePrimarySource()
    subtractive: SubtractivePrimarySource = SubtractivePrimarySource()
    preventive: PreventivePrimarySource = PreventivePrimarySource()


class PrimarySourceRuleSetRequest(_BaseRuleSet):
    """Set of rules to create or update a primary source item."""

    entityType: Annotated[
        Literal["PrimarySourceRuleSetRequest"], Field(alias="$type", frozen=True)
    ] = "PrimarySourceRuleSetRequest"


class PrimarySourceRuleSetResponse(_BaseRuleSet):
    """Set of rules to retrieve a primary source item."""

    entityType: Annotated[
        Literal["PrimarySourceRuleSetResponse"], Field(alias="$type", frozen=True)
    ] = "PrimarySourceRuleSetResponse"
    stableTargetId: MergedPrimarySourceIdentifier


class PrimarySourceMapping(_Stem, BaseMapping):
    """Mapping for describing a primary source transformation."""

    entityType: Annotated[
        Literal["PrimarySourceMapping"], Field(alias="$type", frozen=True)
    ] = "PrimarySourceMapping"
    hadPrimarySource: Annotated[
        list[MappingField[MergedPrimarySourceIdentifier]], Field(min_length=1)
    ]
    identifierInPrimarySource: Annotated[list[MappingField[str]], Field(min_length=1)]
    version: list[MappingField[VersionStr | None]] = []
    alternativeTitle: list[MappingField[list[Text]]] = []
    contact: list[MappingField[list[AnyContactIdentifier]]] = []
    description: list[MappingField[list[Text]]] = []
    documentation: list[MappingField[list[Link]]] = []
    locatedAt: list[MappingField[list[Link]]] = []
    title: list[MappingField[list[Text]]] = []
    unitInCharge: list[MappingField[list[MergedOrganizationalUnitIdentifier]]] = []


class PrimarySourceFilter(_Stem, BaseFilter):
    """Class for defining filter rules for primary source items."""

    entityType: Annotated[
        Literal["PrimarySourceFilter"], Field(alias="$type", frozen=True)
    ] = "PrimarySourceFilter"
    fields: Annotated[list[FilterField], Field(title="fields")] = []



from abc import ABCMeta, abstractmethod
from contextlib import ExitStack, closing
from typing import Self, cast, final

from mex.common.context import SingletonStore
from mex.common.transform import dromedary_to_snake


class _ConnectorStore(SingletonStore["BaseConnector"]):
    """Thin wrapper for storing thread-local singletons of connectors."""

    def reset(self) -> None:
        """Close all connectors and remove them from the singleton store."""
        with ExitStack() as stack:
            for connector in self:
                stack.push(closing(connector))
        super().reset()

    def metrics(self) -> dict[str, int]:
        """Generate metrics about all active connectors."""
        return {
            f"{dromedary_to_snake(connector.__class__.__name__)}_{metric_key}": value
            for connector in self
            for metric_key, value in connector.metrics().items()
        }


CONNECTOR_STORE = _ConnectorStore()


class BaseConnector(metaclass=ABCMeta):
    """Base class for connectors that are handled as singletons."""

    @final
    @classmethod
    def get(cls) -> Self:
        """Get the singleton instance for this class from the store."""
        return cast("Self", CONNECTOR_STORE.load(cls))

    @abstractmethod
    def __init__(self) -> None:  # pragma: no cover
        """Create a new connector instance."""

    def metrics(self) -> dict[str, int]:  # pragma: no cover
        """Generate metrics about connector usage."""
        return {}

    @abstractmethod
    def close(self) -> None:  # pragma: no cover
        """Close the connector's underlying sockets."""



"""These models implement the types defined by the `mex-model` in their various stages.

The current set of entity types includes:

- AccessPlatform
- Activity
- BibliographicResource
- Consent
- ContactPoint
- Distribution
- Organization
- OrganizationalUnit
- Person
- PrimarySource
- Resource
- Variable
- VariableGroup

Each entity type `T` is modelled for the following use cases:

- `BaseT` defines all fields according to `mex-model` except for provenance fields
- `ExtractedT` defines an automatically extracted metadata item including provenance
- `MergedT` defines the result of merging extracted items and rules into a single item
- `PreviewT` defines a preview of a merged item without enforcing cardinality validation

- `AdditiveT` defines a rule to add values to specific fields of a merged item
- `SubtractiveT` defines a rule to subtract (or block) specific values for specific
  fields from contributing to a merged item
- `PreventiveT` defines a rule to prevent (or block) specific primary sources from
  contributing to specific fields of a merged item
- `TRuleSet` classes are used for CRUD operations on a set of three rules

- `TFilter` defines how an entity filter specification should look like
- `TMapping` defines how a raw-data to extracted item mapping should look like

Since these models for different use cases have a lot of overlapping attributes,
we use a number of intermediate private classes to compose the public classes:

- `_Stem` defines a static class attribute `stemType`, e.g. `Person` or `PrimarySource`,
  which is added to all intermediate and exported classes
- `_OptionalLists` defines all fields typed as lists with an arity of 0-n
- `_RequiredLists` defines all fields typed as lists with an arity of 1-n
- `_SparseLists` re-defines all fields from `_RequiredLists` with an arity of 0-n

- `_OptionalValues` defines all fields with optional values (arity of 0-1)
- `_RequiredValues` defines all fields with required values (arity of 1)
- `_SparseValues` re-defines all fields from `_RequiredValues` with an arity of 0-1
- `_VariadicValues` re-defines all fields from `_OptionalValues` and `_RequiredValues`
  as list fields with an arity of 0-n

- `_BaseRuleSet` bundles the additive, subtractive and preventive rules for one type

These private classes are used to compose the public classes like so:

- BaseT: _OptionalLists, _RequiredLists, _OptionalValues, _RequiredValues
- ExtractedT: BaseT, ExtractedData
- MergedT: BaseT, MergedItem
- PreviewT: _OptionalLists, _SparseLists, _OptionalValues, _SparseValues, PreviewItem

- AdditiveT: _OptionalLists, _SparseLists, _OptionalValues, _SparseValues, AdditiveRule
- SubtractiveT: _OptionalLists, _SparseLists, _VariadicValues, SubtractiveRule
- PreventiveT: all fields from BaseT re-typed as MergedPrimarySourceIdentifier
- TRuleSetRequest: bundle of all three rules for one type used to create new rules
- TRuleSetResponse: bundle of all three rules for one type including a `stableTargetId`

- TFilter: a single field containing a list of filter rule definitions
- TMapping: all BaseT fields re-typed as lists of mapping fields with `setValues` type

All models intended for developer consumption have two main classifying class variables:
The frozen `entityType` field is added to classes to help with assigning the correct
class when reading raw JSON entities. Simple duck-typing would not work, because some
entity-types have overlapping attributes, like `Person.email` and `ContactPoint.email`.
See: https://docs.pydantic.dev/latest/concepts/fields/#discriminator
The frozen `stemType` class variable is added to classes to help with knowing which
special-use-case classes are meant for the same type of items.
E.g. `ExtractedPerson`, `MergedPerson` and `PreventivePerson` all share the same
`stemType` of `"Person"`.

In addition to the classes themselves, `mex.common.models` also exposes various
lists of models, lookups by class name and typing for unions of models.
"""

from typing import Annotated, Final, get_args

from pydantic import Field, TypeAdapter

from mex.common.models.access_platform import (
    AccessPlatformFilter,
    AccessPlatformMapping,
    AccessPlatformRuleSetRequest,
    AccessPlatformRuleSetResponse,
    AdditiveAccessPlatform,
    BaseAccessPlatform,
    ExtractedAccessPlatform,
    MergedAccessPlatform,
    PreventiveAccessPlatform,
    PreviewAccessPlatform,
    SubtractiveAccessPlatform,
)
from mex.common.models.activity import (
    ActivityFilter,
    ActivityMapping,
    ActivityRuleSetRequest,
    ActivityRuleSetResponse,
    AdditiveActivity,
    BaseActivity,
    ExtractedActivity,
    MergedActivity,
    PreventiveActivity,
    PreviewActivity,
    SubtractiveActivity,
)
from mex.common.models.base.container import (
    ItemsContainer,
    PaginatedItemsContainer,
)
from mex.common.models.base.extracted_data import ExtractedData
from mex.common.models.base.filter import BaseFilter, FilterField, FilterRule
from mex.common.models.base.mapping import BaseMapping, MappingField, MappingRule
from mex.common.models.base.merged_item import MergedItem
from mex.common.models.base.model import BaseModel
from mex.common.models.base.rules import AdditiveRule, PreventiveRule, SubtractiveRule
from mex.common.models.base.status import Status, VersionStatus
from mex.common.models.bibliographic_resource import (
    AdditiveBibliographicResource,
    BaseBibliographicResource,
    BibliographicResourceFilter,
    BibliographicResourceMapping,
    BibliographicResourceRuleSetRequest,
    BibliographicResourceRuleSetResponse,
    ExtractedBibliographicResource,
    MergedBibliographicResource,
    PreventiveBibliographicResource,
    PreviewBibliographicResource,
    SubtractiveBibliographicResource,
)
from mex.common.models.consent import (
    AdditiveConsent,
    BaseConsent,
    ConsentFilter,
    ConsentMapping,
    ConsentRuleSetRequest,
    ConsentRuleSetResponse,
    ExtractedConsent,
    MergedConsent,
    PreventiveConsent,
    PreviewConsent,
    SubtractiveConsent,
)
from mex.common.models.contact_point import (
    AdditiveContactPoint,
    BaseContactPoint,
    ContactPointFilter,
    ContactPointMapping,
    ContactPointRuleSetRequest,
    ContactPointRuleSetResponse,
    ExtractedContactPoint,
    MergedContactPoint,
    PreventiveContactPoint,
    PreviewContactPoint,
    SubtractiveContactPoint,
)
from mex.common.models.distribution import (
    AdditiveDistribution,
    BaseDistribution,
    DistributionFilter,
    DistributionMapping,
    DistributionRuleSetRequest,
    DistributionRuleSetResponse,
    ExtractedDistribution,
    MergedDistribution,
    PreventiveDistribution,
    PreviewDistribution,
    SubtractiveDistribution,
)
from mex.common.models.organization import (
    AdditiveOrganization,
    BaseOrganization,
    ExtractedOrganization,
    MergedOrganization,
    OrganizationFilter,
    OrganizationMapping,
    OrganizationRuleSetRequest,
    OrganizationRuleSetResponse,
    PreventiveOrganization,
    PreviewOrganization,
    SubtractiveOrganization,
)
from mex.common.models.organizational_unit import (
    AdditiveOrganizationalUnit,
    BaseOrganizationalUnit,
    ExtractedOrganizationalUnit,
    MergedOrganizationalUnit,
    OrganizationalUnitFilter,
    OrganizationalUnitMapping,
    OrganizationalUnitRuleSetRequest,
    OrganizationalUnitRuleSetResponse,
    PreventiveOrganizationalUnit,
    PreviewOrganizationalUnit,
    SubtractiveOrganizationalUnit,
)
from mex.common.models.person import (
    AdditivePerson,
    BasePerson,
    ExtractedPerson,
    MergedPerson,
    PersonFilter,
    PersonMapping,
    PersonRuleSetRequest,
    PersonRuleSetResponse,
    PreventivePerson,
    PreviewPerson,
    SubtractivePerson,
)
from mex.common.models.primary_source import (
    AdditivePrimarySource,
    BasePrimarySource,
    ExtractedPrimarySource,
    MergedPrimarySource,
    PreventivePrimarySource,
    PreviewPrimarySource,
    PrimarySourceFilter,
    PrimarySourceMapping,
    PrimarySourceRuleSetRequest,
    PrimarySourceRuleSetResponse,
    SubtractivePrimarySource,
)
from mex.common.models.resource import (
    AdditiveResource,
    BaseResource,
    ExtractedResource,
    MergedResource,
    PreventiveResource,
    PreviewResource,
    ResourceFilter,
    ResourceMapping,
    ResourceRuleSetRequest,
    ResourceRuleSetResponse,
    SubtractiveResource,
)
from mex.common.models.variable import (
    AdditiveVariable,
    BaseVariable,
    ExtractedVariable,
    MergedVariable,
    PreventiveVariable,
    PreviewVariable,
    SubtractiveVariable,
    VariableFilter,
    VariableMapping,
    VariableRuleSetRequest,
    VariableRuleSetResponse,
)
from mex.common.models.variable_group import (
    AdditiveVariableGroup,
    BaseVariableGroup,
    ExtractedVariableGroup,
    MergedVariableGroup,
    PreventiveVariableGroup,
    PreviewVariableGroup,
    SubtractiveVariableGroup,
    VariableGroupFilter,
    VariableGroupMapping,
    VariableGroupRuleSetRequest,
    VariableGroupRuleSetResponse,
)
from mex.common.types import (
    ExtractedPrimarySourceIdentifier,
    MergedPrimarySourceIdentifier,
)

__all__ = (
    "ADDITIVE_MODEL_CLASSES",
    "ADDITIVE_MODEL_CLASSES_BY_NAME",
    "BASE_MODEL_CLASSES",
    "BASE_MODEL_CLASSES_BY_NAME",
    "EXTRACTED_MODEL_CLASSES",
    "EXTRACTED_MODEL_CLASSES_BY_NAME",
    "FILTER_MODEL_BY_EXTRACTED_CLASS_NAME",
    "MAPPING_MODEL_BY_EXTRACTED_CLASS_NAME",
    "MERGED_MODEL_CLASSES",
    "MERGED_MODEL_CLASSES_BY_NAME",
    "MEX_PRIMARY_SOURCE_IDENTIFIER",
    "MEX_PRIMARY_SOURCE_IDENTIFIER_IN_PRIMARY_SOURCE",
    "MEX_PRIMARY_SOURCE_STABLE_TARGET_ID",
    "PREVENTIVE_MODEL_CLASSES",
    "PREVENTIVE_MODEL_CLASSES_BY_NAME",
    "RULE_MODEL_CLASSES",
    "RULE_MODEL_CLASSES_BY_NAME",
    "RULE_SET_REQUEST_CLASSES",
    "RULE_SET_REQUEST_CLASSES_BY_NAME",
    "RULE_SET_RESPONSE_CLASSES",
    "RULE_SET_RESPONSE_CLASSES_BY_NAME",
    "SUBTRACTIVE_MODEL_CLASSES",
    "SUBTRACTIVE_MODEL_CLASSES_BY_NAME",
    "AccessPlatformFilter",
    "AccessPlatformMapping",
    "AccessPlatformRuleSetRequest",
    "AccessPlatformRuleSetResponse",
    "ActivityFilter",
    "ActivityMapping",
    "ActivityRuleSetRequest",
    "ActivityRuleSetResponse",
    "AdditiveAccessPlatform",
    "AdditiveActivity",
    "AdditiveBibliographicResource",
    "AdditiveConsent",
    "AdditiveContactPoint",
    "AdditiveDistribution",
    "AdditiveModelTypeAdapter",
    "AdditiveOrganization",
    "AdditiveOrganizationalUnit",
    "AdditivePerson",
    "AdditivePrimarySource",
    "AdditiveResource",
    "AdditiveRule",
    "AdditiveVariable",
    "AdditiveVariableGroup",
    "AnyAdditiveModel",
    "AnyBaseModel",
    "AnyExtractedModel",
    "AnyMergedModel",
    "AnyPreventiveModel",
    "AnyPreviewModel",
    "AnyRuleModel",
    "AnyRuleSetRequest",
    "AnyRuleSetResponse",
    "AnySubtractiveModel",
    "BaseAccessPlatform",
    "BaseActivity",
    "BaseBibliographicResource",
    "BaseConsent",
    "BaseContactPoint",
    "BaseDistribution",
    "BaseFilter",
    "BaseMapping",
    "BaseModel",
    "BaseOrganization",
    "BaseOrganizationalUnit",
    "BasePerson",
    "BasePrimarySource",
    "BaseResource",
    "BaseVariable",
    "BaseVariableGroup",
    "BibliographicResourceFilter",
    "BibliographicResourceMapping",
    "ConsentFilter",
    "ConsentMapping",
    "ConsentRuleSetRequest",
    "ConsentRuleSetResponse",
    "ContactPointFilter",
    "ContactPointMapping",
    "ContactPointRuleSetRequest",
    "ContactPointRuleSetResponse",
    "DistributionFilter",
    "DistributionMapping",
    "DistributionRuleSetRequest",
    "DistributionRuleSetResponse",
    "ExtractedAccessPlatform",
    "ExtractedActivity",
    "ExtractedBibliographicResource",
    "ExtractedConsent",
    "ExtractedContactPoint",
    "ExtractedData",
    "ExtractedDistribution",
    "ExtractedModelTypeAdapter",
    "ExtractedOrganization",
    "ExtractedOrganizationalUnit",
    "ExtractedPerson",
    "ExtractedPrimarySource",
    "ExtractedResource",
    "ExtractedVariable",
    "ExtractedVariableGroup",
    "FilterField",
    "FilterModelTypeAdapter",
    "FilterRule",
    "ItemsContainer",
    "MappingField",
    "MappingModelTypeAdapter",
    "MappingRule",
    "MergedAccessPlatform",
    "MergedActivity",
    "MergedBibliographicResource",
    "MergedConsent",
    "MergedContactPoint",
    "MergedDistribution",
    "MergedItem",
    "MergedModelTypeAdapter",
    "MergedOrganization",
    "MergedOrganizationalUnit",
    "MergedPerson",
    "MergedPrimarySource",
    "MergedResource",
    "MergedVariable",
    "MergedVariableGroup",
    "OrganizationFilter",
    "OrganizationMapping",
    "OrganizationRuleSetRequest",
    "OrganizationRuleSetResponse",
    "OrganizationalUnitFilter",
    "OrganizationalUnitMapping",
    "OrganizationalUnitRuleSetRequest",
    "OrganizationalUnitRuleSetResponse",
    "PaginatedItemsContainer",
    "PersonFilter",
    "PersonMapping",
    "PersonRuleSetRequest",
    "PersonRuleSetResponse",
    "PreventiveAccessPlatform",
    "PreventiveActivity",
    "PreventiveBibliographicResource",
    "PreventiveConsent",
    "PreventiveContactPoint",
    "PreventiveDistribution",
    "PreventiveModelTypeAdapter",
    "PreventiveOrganization",
    "PreventiveOrganizationalUnit",
    "PreventivePerson",
    "PreventivePrimarySource",
    "PreventiveResource",
    "PreventiveRule",
    "PreventiveVariable",
    "PreventiveVariableGroup",
    "PreviewAccessPlatform",
    "PreviewActivity",
    "PreviewBibliographicResource",
    "PreviewConsent",
    "PreviewContactPoint",
    "PreviewDistribution",
    "PreviewModelTypeAdapter",
    "PreviewOrganization",
    "PreviewOrganizationalUnit",
    "PreviewPerson",
    "PreviewPrimarySource",
    "PreviewResource",
    "PreviewVariable",
    "PreviewVariableGroup",
    "PrimarySourceFilter",
    "PrimarySourceMapping",
    "PrimarySourceRuleSetRequest",
    "PrimarySourceRuleSetResponse",
    "ResourceFilter",
    "ResourceMapping",
    "ResourceRuleSetRequest",
    "ResourceRuleSetResponse",
    "RuleModelTypeAdapter",
    "RuleSetRequestTypeAdapter",
    "RuleSetResponseTypeAdapter",
    "Status",
    "SubtractiveAccessPlatform",
    "SubtractiveActivity",
    "SubtractiveBibliographicResource",
    "SubtractiveConsent",
    "SubtractiveContactPoint",
    "SubtractiveDistribution",
    "SubtractiveModelTypeAdapter",
    "SubtractiveOrganization",
    "SubtractiveOrganizationalUnit",
    "SubtractivePerson",
    "SubtractivePrimarySource",
    "SubtractiveResource",
    "SubtractiveRule",
    "SubtractiveVariable",
    "SubtractiveVariableGroup",
    "VariableFilter",
    "VariableGroupFilter",
    "VariableGroupMapping",
    "VariableGroupRuleSetRequest",
    "VariableGroupRuleSetResponse",
    "VariableMapping",
    "VariableRuleSetRequest",
    "VariableRuleSetResponse",
    "VersionStatus",
)

MEX_PRIMARY_SOURCE_IDENTIFIER = ExtractedPrimarySourceIdentifier("00000000000001")
MEX_PRIMARY_SOURCE_IDENTIFIER_IN_PRIMARY_SOURCE = "mex"
MEX_PRIMARY_SOURCE_STABLE_TARGET_ID = MergedPrimarySourceIdentifier("00000000000000")

AnyBaseModel = (
    BaseAccessPlatform
    | BaseActivity
    | BaseBibliographicResource
    | BaseConsent
    | BaseContactPoint
    | BaseDistribution
    | BaseOrganization
    | BaseOrganizationalUnit
    | BasePerson
    | BasePrimarySource
    | BaseResource
    | BaseVariable
    | BaseVariableGroup
)
BASE_MODEL_CLASSES: Final[list[type[AnyBaseModel]]] = list(get_args(AnyBaseModel))
BASE_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnyBaseModel]]] = {
    cls.__name__: cls for cls in BASE_MODEL_CLASSES
}

AnyExtractedModel = (
    ExtractedAccessPlatform
    | ExtractedActivity
    | ExtractedBibliographicResource
    | ExtractedConsent
    | ExtractedContactPoint
    | ExtractedDistribution
    | ExtractedOrganization
    | ExtractedOrganizationalUnit
    | ExtractedPerson
    | ExtractedPrimarySource
    | ExtractedResource
    | ExtractedVariable
    | ExtractedVariableGroup
)
ExtractedModelTypeAdapter: TypeAdapter[AnyExtractedModel] = TypeAdapter(
    Annotated[AnyExtractedModel, Field(discriminator="entityType")]
)
EXTRACTED_MODEL_CLASSES: Final[list[type[AnyExtractedModel]]] = list(
    get_args(AnyExtractedModel)
)
EXTRACTED_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnyExtractedModel]]] = {
    cls.__name__: cls for cls in EXTRACTED_MODEL_CLASSES
}

AnyMergedModel = (
    MergedAccessPlatform
    | MergedActivity
    | MergedBibliographicResource
    | MergedConsent
    | MergedContactPoint
    | MergedDistribution
    | MergedOrganization
    | MergedOrganizationalUnit
    | MergedPerson
    | MergedPrimarySource
    | MergedResource
    | MergedVariable
    | MergedVariableGroup
)
MergedModelTypeAdapter: TypeAdapter[AnyMergedModel] = TypeAdapter(
    Annotated[AnyMergedModel, Field(discriminator="entityType")]
)
MERGED_MODEL_CLASSES: Final[list[type[AnyMergedModel]]] = list(get_args(AnyMergedModel))
MERGED_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnyMergedModel]]] = {
    cls.__name__: cls for cls in MERGED_MODEL_CLASSES
}

AnyPreviewModel = (
    PreviewAccessPlatform
    | PreviewActivity
    | PreviewBibliographicResource
    | PreviewConsent
    | PreviewContactPoint
    | PreviewDistribution
    | PreviewOrganization
    | PreviewOrganizationalUnit
    | PreviewPerson
    | PreviewPrimarySource
    | PreviewResource
    | PreviewVariable
    | PreviewVariableGroup
)
PreviewModelTypeAdapter: TypeAdapter[AnyPreviewModel] = TypeAdapter(
    Annotated[AnyPreviewModel, Field(discriminator="entityType")]
)
PREVIEW_MODEL_CLASSES: Final[list[type[AnyPreviewModel]]] = list(
    get_args(AnyPreviewModel)
)
PREVIEW_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnyPreviewModel]]] = {
    cls.__name__: cls for cls in PREVIEW_MODEL_CLASSES
}

AnyAdditiveModel = (
    AdditiveAccessPlatform
    | AdditiveActivity
    | AdditiveBibliographicResource
    | AdditiveConsent
    | AdditiveContactPoint
    | AdditiveDistribution
    | AdditiveOrganization
    | AdditiveOrganizationalUnit
    | AdditivePerson
    | AdditivePrimarySource
    | AdditiveResource
    | AdditiveVariable
    | AdditiveVariableGroup
)
AdditiveModelTypeAdapter: TypeAdapter[AnyAdditiveModel] = TypeAdapter(
    Annotated[AnyAdditiveModel, Field(discriminator="entityType")]
)
ADDITIVE_MODEL_CLASSES: Final[list[type[AnyAdditiveModel]]] = list(
    get_args(AnyAdditiveModel)
)
ADDITIVE_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnyAdditiveModel]]] = {
    cls.__name__: cls for cls in ADDITIVE_MODEL_CLASSES
}

AnySubtractiveModel = (
    SubtractiveAccessPlatform
    | SubtractiveActivity
    | SubtractiveBibliographicResource
    | SubtractiveConsent
    | SubtractiveContactPoint
    | SubtractiveDistribution
    | SubtractiveOrganization
    | SubtractiveOrganizationalUnit
    | SubtractivePerson
    | SubtractivePrimarySource
    | SubtractiveResource
    | SubtractiveVariable
    | SubtractiveVariableGroup
)
SubtractiveModelTypeAdapter: TypeAdapter[AnySubtractiveModel] = TypeAdapter(
    Annotated[AnySubtractiveModel, Field(discriminator="entityType")]
)
SUBTRACTIVE_MODEL_CLASSES: Final[list[type[AnySubtractiveModel]]] = list(
    get_args(AnySubtractiveModel)
)
SUBTRACTIVE_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnySubtractiveModel]]] = {
    cls.__name__: cls for cls in SUBTRACTIVE_MODEL_CLASSES
}

AnyPreventiveModel = (
    PreventiveAccessPlatform
    | PreventiveActivity
    | PreventiveBibliographicResource
    | PreventiveConsent
    | PreventiveContactPoint
    | PreventiveDistribution
    | PreventiveOrganization
    | PreventiveOrganizationalUnit
    | PreventivePerson
    | PreventivePrimarySource
    | PreventiveResource
    | PreventiveVariable
    | PreventiveVariableGroup
)
PreventiveModelTypeAdapter: TypeAdapter[AnyPreventiveModel] = TypeAdapter(
    Annotated[AnyPreventiveModel, Field(discriminator="entityType")]
)
PREVENTIVE_MODEL_CLASSES: Final[list[type[AnyPreventiveModel]]] = list(
    get_args(AnyPreventiveModel)
)
PREVENTIVE_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnyPreventiveModel]]] = {
    cls.__name__: cls for cls in PREVENTIVE_MODEL_CLASSES
}

AnyRuleModel = AnyAdditiveModel | AnySubtractiveModel | AnyPreventiveModel
RuleModelTypeAdapter: TypeAdapter[AnyRuleModel] = TypeAdapter(
    Annotated[AnyRuleModel, Field(discriminator="entityType")]
)
RULE_MODEL_CLASSES: Final[list[type[AnyRuleModel]]] = list(get_args(AnyRuleModel))
RULE_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnyRuleModel]]] = {
    cls.__name__: cls for cls in RULE_MODEL_CLASSES
}

AnyRuleSetRequest = (
    AccessPlatformRuleSetRequest
    | ActivityRuleSetRequest
    | BibliographicResourceRuleSetRequest
    | ConsentRuleSetRequest
    | ContactPointRuleSetRequest
    | DistributionRuleSetRequest
    | OrganizationRuleSetRequest
    | OrganizationalUnitRuleSetRequest
    | PersonRuleSetRequest
    | PrimarySourceRuleSetRequest
    | ResourceRuleSetRequest
    | VariableRuleSetRequest
    | VariableGroupRuleSetRequest
)
RuleSetRequestTypeAdapter: TypeAdapter[AnyRuleSetRequest] = TypeAdapter(
    Annotated[AnyRuleSetRequest, Field(discriminator="entityType")]
)
RULE_SET_REQUEST_CLASSES: Final[list[type[AnyRuleSetRequest]]] = list(
    get_args(AnyRuleSetRequest)
)
RULE_SET_REQUEST_CLASSES_BY_NAME: Final[dict[str, type[AnyRuleSetRequest]]] = {
    cls.__name__: cls for cls in RULE_SET_REQUEST_CLASSES
}

AnyRuleSetResponse = (
    AccessPlatformRuleSetResponse
    | ActivityRuleSetResponse
    | BibliographicResourceRuleSetResponse
    | ConsentRuleSetResponse
    | ContactPointRuleSetResponse
    | DistributionRuleSetResponse
    | OrganizationRuleSetResponse
    | OrganizationalUnitRuleSetResponse
    | PersonRuleSetResponse
    | PrimarySourceRuleSetResponse
    | ResourceRuleSetResponse
    | VariableRuleSetResponse
    | VariableGroupRuleSetResponse
)
RuleSetResponseTypeAdapter: TypeAdapter[AnyRuleSetResponse] = TypeAdapter(
    Annotated[AnyRuleSetResponse, Field(discriminator="entityType")]
)
RULE_SET_RESPONSE_CLASSES: Final[list[type[AnyRuleSetResponse]]] = list(
    get_args(AnyRuleSetResponse)
)
RULE_SET_RESPONSE_CLASSES_BY_NAME: Final[dict[str, type[AnyRuleSetResponse]]] = {
    cls.__name__: cls for cls in RULE_SET_RESPONSE_CLASSES
}

AnyMappingModel = (
    AccessPlatformMapping
    | ActivityMapping
    | BibliographicResourceMapping
    | ConsentMapping
    | ContactPointMapping
    | DistributionMapping
    | OrganizationMapping
    | OrganizationalUnitMapping
    | PersonMapping
    | PrimarySourceMapping
    | ResourceMapping
    | VariableMapping
    | VariableGroupMapping
)
MappingModelTypeAdapter: TypeAdapter[AnyMappingModel] = TypeAdapter(
    Annotated[AnyMappingModel, Field(discriminator="entityType")]
)
MAPPING_MODEL_CLASSES: Final[list[type[AnyMappingModel]]] = list(
    get_args(AnyMappingModel)
)
MAPPING_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnyMappingModel]]] = {
    cls.__name__: cls for cls in MAPPING_MODEL_CLASSES
}
# MAPPING_MODEL_BY_EXTRACTED_CLASS_NAME is deprecated, use MAPPING_MODEL_CLASSES_BY_NAME
MAPPING_MODEL_BY_EXTRACTED_CLASS_NAME = {
    f"Extracted{cls.stemType}": cls for cls in MAPPING_MODEL_CLASSES
}

AnyFilterModel = (
    AccessPlatformFilter
    | ActivityFilter
    | BibliographicResourceFilter
    | ConsentFilter
    | ContactPointFilter
    | DistributionFilter
    | OrganizationFilter
    | OrganizationalUnitFilter
    | PersonFilter
    | PrimarySourceFilter
    | ResourceFilter
    | VariableFilter
    | VariableGroupFilter
)
FilterModelTypeAdapter: TypeAdapter[AnyFilterModel] = TypeAdapter(
    Annotated[AnyFilterModel, Field(discriminator="entityType")]
)
FILTER_MODEL_CLASSES: Final[list[type[AnyFilterModel]]] = list(get_args(AnyFilterModel))
FILTER_MODEL_CLASSES_BY_NAME: Final[dict[str, type[AnyFilterModel]]] = {
    cls.__name__: cls for cls in FILTER_MODEL_CLASSES
}
# FILTER_MODEL_BY_EXTRACTED_CLASS_NAME is deprecated, use FILTER_MODEL_CLASSES_BY_NAME
FILTER_MODEL_BY_EXTRACTED_CLASS_NAME = {
    f"Extracted{cls.stemType}": cls for cls in FILTER_MODEL_CLASSES
}



from collections.abc import Generator

import pytest

from mex.common.identity.memory import MemoryIdentityProvider
from mex.common.identity.registry import _PROVIDER_REGISTRY, register_provider
from mex.common.types import IdentityProvider


@pytest.fixture(autouse=True)
def isolate_provider_registry() -> Generator[None, None, None]:
    """Restore the provider registry after each identity test."""
    yield
    _PROVIDER_REGISTRY.clear()
    register_provider(IdentityProvider.MEMORY, MemoryIdentityProvider)
